from unittest import (
    TestCase,
    mock,
)

from pcs import settings
from pcs.common.reports import codes as report_codes
from pcs.lib.external import CommandRunner
from pcs.lib.resource_agent import list as ra_list
from pcs.lib.resource_agent.error import (
    AgentNameGuessFoundMoreThanOne,
    AgentNameGuessFoundNone,
)
from pcs.lib.resource_agent.types import (
    ResourceAgentName,
    StandardProviderTuple,
)

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


class ListResourceAgentsStandards(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)

    def assert_runner(self):
        self.mock_runner.run.assert_called_once_with(
            [settings.crm_resource_binary, "--list-standards"]
        )

    def test_success(self):
        agents = [
            "",
            "service  ",
            "systemd",
            "  nagios  ",
            "ocf",
            "  lsb",
            "  service",
            "",
            "stonith",
            "",
        ]
        # retval is number of standards found
        self.mock_runner.run.return_value = (
            "\n".join(agents) + "\n",
            "",
            len(agents),
        )

        self.assertEqual(
            ra_list.list_resource_agents_standards(self.mock_runner),
            ["lsb", "nagios", "ocf", "service", "stonith", "systemd"],
        )
        self.assert_runner()

    def test_empty(self):
        self.mock_runner.run.return_value = ("", "", 0)
        self.assertEqual(
            ra_list.list_resource_agents_standards(self.mock_runner), []
        )
        self.assert_runner()

    def test_error(self):
        self.mock_runner.run.return_value = ("lsb", "error", 1)
        self.assertEqual(
            ra_list.list_resource_agents_standards(self.mock_runner), ["lsb"]
        )
        self.assert_runner()


class ListResourceAgentsOcfProviders(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)

    def assert_runner(self):
        self.mock_runner.run.assert_called_once_with(
            [settings.crm_resource_binary, "--list-ocf-providers"]
        )

    def test_success_filter_whitespace(self):
        providers = [
            "",
            "heartbeat",
            "  pacemaker",
            " openstack",
            "pacemaker ",
            " booth ",
        ]
        # retval is number of providers found
        self.mock_runner.run.return_value = (
            "\n".join(providers) + "\n",
            "",
            len(providers),
        )

        self.assertEqual(
            ra_list.list_resource_agents_ocf_providers(self.mock_runner),
            ["booth", "heartbeat", "openstack", "pacemaker"],
        )
        self.assert_runner()

    def test_empty(self):
        self.mock_runner.run.return_value = ("", "", 0)
        self.assertEqual(
            ra_list.list_resource_agents_ocf_providers(self.mock_runner), []
        )
        self.assert_runner()

    def test_error(self):
        self.mock_runner.run.return_value = ("booth", "error", 1)
        self.assertEqual(
            ra_list.list_resource_agents_ocf_providers(self.mock_runner),
            ["booth"],
        )
        self.assert_runner()


class ListResourceAgentsStandardsAndProviders(TestCase):
    def test_success(self):
        mock_runner = mock.MagicMock(spec_set=CommandRunner)
        mock_runner.run.side_effect = [
            (
                "\n".join(
                    [
                        "nagios ",
                        "ocf",
                        " service",
                        "lsb  ",
                        " ocf",
                        "stonith",
                        "systemd",
                        "",
                    ]
                ),
                "",
                0,
            ),
            (
                "\n".join(
                    [
                        "heartbeat",
                        "openstack",
                        "pacemaker",
                        "booth ",
                        " pacemaker",
                        "",
                    ]
                ),
                "",
                0,
            ),
        ]

        self.assertEqual(
            ra_list.list_resource_agents_standards_and_providers(mock_runner),
            [
                StandardProviderTuple("lsb"),
                StandardProviderTuple("nagios"),
                StandardProviderTuple("ocf", "booth"),
                StandardProviderTuple("ocf", "heartbeat"),
                StandardProviderTuple("ocf", "openstack"),
                StandardProviderTuple("ocf", "pacemaker"),
                StandardProviderTuple("service"),
                StandardProviderTuple("stonith"),
                StandardProviderTuple("systemd"),
            ],
        )

        self.assertEqual(2, len(mock_runner.run.mock_calls))
        mock_runner.run.assert_has_calls(
            [
                mock.call([settings.crm_resource_binary, "--list-standards"]),
                mock.call(
                    [
                        settings.crm_resource_binary,
                        "--list-ocf-providers",
                    ]
                ),
            ]
        )


class ListResourceAgents(TestCase):
    def setUp(self):
        self.mock_runner = mock.MagicMock(spec_set=CommandRunner)

    def assert_runner(self, standard_provider):
        self.mock_runner.run.assert_called_once_with(
            [settings.crm_resource_binary, "--list-agents", standard_provider]
        )

    def test_success_standard(self):
        self.mock_runner.run.return_value = (
            "\n".join(
                [
                    "docker",
                    "Dummy",
                    "dhcpd",
                    "Dummy",
                    "ethmonitor",
                    "",
                ]
            ),
            "",
            0,
        )

        self.assertEqual(
            ra_list.list_resource_agents(
                self.mock_runner, StandardProviderTuple("ocf")
            ),
            [
                "dhcpd",
                "docker",
                "Dummy",
                "ethmonitor",
            ],
        )
        self.assert_runner("ocf")

    def test_success_standard_provider(self):
        self.mock_runner.run.return_value = (
            "\n".join(
                [
                    "ping",
                    "SystemHealth",
                    "SysInfo",
                    "HealthCPU",
                    "Dummy",
                    "",
                ]
            ),
            "",
            0,
        )

        self.assertEqual(
            ra_list.list_resource_agents(
                self.mock_runner, StandardProviderTuple("ocf", "pacemaker")
            ),
            [
                "Dummy",
                "HealthCPU",
                "ping",
                "SysInfo",
                "SystemHealth",
            ],
        )
        self.assert_runner("ocf:pacemaker")

    def test_bad_standard(self):
        self.mock_runner.run.return_value = (
            "",
            "No agents found for standard=nonsense, provider=*",
            1,
        )
        self.assertEqual(
            ra_list.list_resource_agents(
                self.mock_runner, StandardProviderTuple("nonsense")
            ),
            [],
        )
        self.assert_runner("nonsense")

    def test_filter_hidden_agents(self):
        self.mock_runner.run.return_value = (
            "\n".join(
                [
                    "fence_na",
                    "fence_wti",
                    "fence_scsi",
                    "fence_vmware_helper",
                    "fence_nss_wrapper",
                    "fence_node",
                    "fence_vmware_soap",
                    "fence_virt",
                    "fence_pcmk",
                    "fence_sanlockd",
                    "fence_xvm",
                    "fence_ack_manual",
                    "fence_legacy",
                    "fence_check",
                    "fence_tool",
                    "fence_kdump_send",
                    "fence_virtd",
                    "",
                ]
            ),
            "",
            0,
        )

        self.assertEqual(
            ra_list.list_resource_agents(
                self.mock_runner, StandardProviderTuple("stonith")
            ),
            [
                "fence_scsi",
                "fence_virt",
                "fence_vmware_soap",
                "fence_wti",
                "fence_xvm",
            ],
        )
        self.assert_runner("stonith")


class FindOneResourceAgentByType(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def _fixture_success_calls(self):
        self.config.runner.pcmk.list_agents_standards(
            "\n".join(["service", "ocf"])
        )
        self.config.runner.pcmk.list_agents_ocf_providers(
            "\n".join(["heartbeat", "pacemaker"])
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:heartbeat",
            "\n".join(["Delay", "Dummy"]),
            name="runner.pcmk.list_agents_ocf_providers.heartbeat",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:pacemaker",
            "\n".join(["Dummy", "Stateful"]),
            name="runner.pcmk.list_agents_ocf_providers.pacemaker",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "service",
            "\n".join(["sshd", "nonvalid:dummy"]),
            name="runner.pcmk.list_agents_ocf_providers.service",
        )

    def test_found_one_agent(self):
        self._fixture_success_calls()
        env = self.env_assist.get_env()
        self.assertEqual(
            ra_list.find_one_resource_agent_by_type(
                env.cmd_runner(), env.report_processor, "delay"
            ),
            ResourceAgentName("ocf", "heartbeat", "Delay"),
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    report_codes.AGENT_NAME_GUESSED,
                    entered_name="delay",
                    guessed_name="ocf:heartbeat:Delay",
                )
            ]
        )

    def test_found_more_agents(self):
        self._fixture_success_calls()
        env = self.env_assist.get_env()
        with self.assertRaises(AgentNameGuessFoundMoreThanOne) as cm:
            ra_list.find_one_resource_agent_by_type(
                env.cmd_runner(), env.report_processor, "dummy"
            )
        self.assertEqual(cm.exception.agent_name, "dummy")
        self.assertEqual(
            cm.exception.names_found,
            ["ocf:heartbeat:Dummy", "ocf:pacemaker:Dummy"],
        )

    def test_found_no_agents(self):
        self._fixture_success_calls()
        env = self.env_assist.get_env()
        with self.assertRaises(AgentNameGuessFoundNone) as cm:
            ra_list.find_one_resource_agent_by_type(
                env.cmd_runner(), env.report_processor, "missing"
            )
        self.assertEqual(cm.exception.agent_name, "missing")



from pcs import (
    status,
    usage,
)
from pcs.cli.booth.command import status as booth_status_cmd
from pcs.cli.common.errors import raise_command_replaced
from pcs.cli.common.routing import create_router
from pcs.pcsd import pcsd_status_cmd
from pcs.qdevice import qdevice_status_cmd
from pcs.quorum import quorum_status_cmd
from pcs.resource import resource_status

status_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: usage.status(argv),
        "booth": booth_status_cmd,
        "corosync": status.corosync_status,
        "cluster": status.cluster_status,
        "nodes": status.nodes_status,
        "pcsd": pcsd_status_cmd,
        "qdevice": qdevice_status_cmd,
        "quorum": quorum_status_cmd,
        "resources": resource_status,
        "xml": status.xml_status,
        "status": status.full_status,
        # removed commands
        # These print error messages which point users to the changes section in
        # pcs manpage.
        # To be removed in the next significant version.
        "groups": lambda lib, argv, modifiers: raise_command_replaced(
            "pcs resource group list"
        ),
    },
    ["status"],
    default_cmd="status",
)



from unittest import (
    TestCase,
    mock,
)

import pcs.lib.commands.sbd as cmd_sbd
from pcs import settings
from pcs.common import reports
from pcs.common.node_communicator import RequestTarget
from pcs.common.reports import ReportItem
from pcs.common.reports import ReportItemSeverity as Severities
from pcs.lib.errors import LibraryError

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_report_item_list_equal,
)
from pcs_test.tools.command_env import get_env_tools
from pcs_test.tools.misc import outdent

# pylint: disable=protected-access


def _assert_equal_list_of_dictionaries_without_order(expected, actual):
    for item in actual:
        if item not in expected:
            raise AssertionError("Given but not expected: {0}".format(item))
    for item in expected:
        if item not in actual:
            raise AssertionError("Expected but not given: {0}".format(item))


class ValidateSbdOptionsTest(TestCase):
    def setUp(self):
        self.allowed_sbd_options = sorted(
            [
                "SBD_DELAY_START",
                "SBD_STARTMODE",
                "SBD_WATCHDOG_TIMEOUT",
                "SBD_TIMEOUT_ACTION",
            ]
        )
        self.timeout_action_allowed_values = sorted(
            [
                "flush",
                "noflush",
                "off",
                "reboot",
                "crashdump",
                "flush,off",
                "flush,reboot",
                "flush,crashdump",
                "noflush,off",
                "noflush,reboot",
                "noflush,crashdump",
                "off,flush",
                "off,noflush",
                "reboot,flush",
                "reboot,noflush",
                "crashdump,flush",
                "crashdump,noflush",
            ]
        )

    def test_all_ok(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean",
            "SBD_TIMEOUT_ACTION": "flush,reboot",
        }
        self.assertEqual([], cmd_sbd._validate_sbd_options(config))

    def test_invalid_value(self):
        config = {
            "SBD_TIMEOUT_ACTION": "flush,noflush",
        }
        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    force_code=reports.codes.FORCE,
                    option_name="SBD_TIMEOUT_ACTION",
                    option_value="flush,noflush",
                    allowed_values=self.timeout_action_allowed_values,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_invalid_value_forced(self):
        config = {
            "SBD_TIMEOUT_ACTION": "flush,noflush",
        }
        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(
                config, allow_invalid_option_values=True
            ),
            [
                fixture.warn(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="SBD_TIMEOUT_ACTION",
                    option_value="flush,noflush",
                    allowed_values=self.timeout_action_allowed_values,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_unknown_options(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean",
            "SBD_UNKNOWN": "",
            "another_unknown_option": "some value",
        }
        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=sorted(
                        ["SBD_UNKNOWN", "another_unknown_option"]
                    ),
                    option_type=None,
                    allowed=self.allowed_sbd_options,
                    allowed_patterns=[],
                    force_code=reports.codes.FORCE,
                ),
            ],
        )

    def test_unknown_options_forced(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean",
            "SBD_UNKNOWN": "",
            "another_unknown_option": "some value",
        }
        # just make sure there is no exception raised
        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config, allow_unknown_opts=True),
            [
                fixture.warn(
                    reports.codes.INVALID_OPTIONS,
                    option_names=sorted(
                        ["SBD_UNKNOWN", "another_unknown_option"]
                    ),
                    option_type=None,
                    allowed=self.allowed_sbd_options,
                    allowed_patterns=[],
                ),
            ],
        )

    def test_unsupported_options(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean",
            "SBD_WATCHDOG_DEV": "/dev/watchdog",
            "SBD_OPTS": "  ",
            "SBD_DEVICE": "/dev/vda",
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=sorted(
                        ["SBD_WATCHDOG_DEV", "SBD_OPTS", "SBD_DEVICE"]
                    ),
                    option_type=None,
                    allowed=self.allowed_sbd_options,
                    allowed_patterns=[],
                ),
            ],
        )

    def test_invalid_and_unsupported_options(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean",
            "SBD_WATCHDOG_DEV": "/dev/watchdog",
            "SBD_UNKNOWN": "",
            "SBD_OPTS": "  ",
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=sorted(["SBD_WATCHDOG_DEV", "SBD_OPTS"]),
                    option_type=None,
                    allowed=self.allowed_sbd_options,
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["SBD_UNKNOWN"],
                    option_type=None,
                    allowed=self.allowed_sbd_options,
                    allowed_patterns=[],
                    force_code=reports.codes.FORCE,
                ),
            ],
        )

    def test_invalid_and_unsupported_options_forced(self):
        config = {
            "SBD_DELAY_START": "yes",
            "SBD_WATCHDOG_TIMEOUT": "5",
            "SBD_STARTMODE": "clean",
            "SBD_WATCHDOG_DEV": "/dev/watchdog",
            "SBD_UNKNOWN": "",
            "SBD_OPTS": "  ",
            "SBD_PACEMAKER": "false",
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config, allow_unknown_opts=True),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=sorted(
                        ["SBD_WATCHDOG_DEV", "SBD_OPTS", "SBD_PACEMAKER"]
                    ),
                    option_type=None,
                    allowed=self.allowed_sbd_options,
                    allowed_patterns=[],
                ),
                fixture.warn(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["SBD_UNKNOWN"],
                    option_type=None,
                    allowed=self.allowed_sbd_options,
                    allowed_patterns=[],
                ),
            ],
        )

    def test_watchdog_timeout_is_not_present(self):
        config = {"SBD_DELAY_START": "yes", "SBD_STARTMODE": "clean"}
        self.assertEqual([], cmd_sbd._validate_sbd_options(config))

    def test_watchdog_timeout_is_nonnegative_int(self):
        # pylint: disable=no-self-use
        config = {
            "SBD_WATCHDOG_TIMEOUT": "-1",
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="SBD_WATCHDOG_TIMEOUT",
                    option_value="-1",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )

    def test_watchdog_timeout_is_not_int(self):
        # pylint: disable=no-self-use
        config = {
            "SBD_WATCHDOG_TIMEOUT": "not int",
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="SBD_WATCHDOG_TIMEOUT",
                    option_value="not int",
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )

    def test_watchdog_timeout_is_none(self):
        # pylint: disable=no-self-use
        config = {
            "SBD_WATCHDOG_TIMEOUT": None,
        }

        assert_report_item_list_equal(
            cmd_sbd._validate_sbd_options(config),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="SBD_WATCHDOG_TIMEOUT",
                    option_value=None,
                    allowed_values="a non-negative integer",
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )


class ValidateWatchdogDictTest(TestCase):
    def test_all_ok(self):
        watchdog_dict = {
            "node1": "/dev/watchdog1",
            "node2": "/dev/watchdog2",
        }
        self.assertEqual([], cmd_sbd._validate_watchdog_dict(watchdog_dict))

    def test_some_not_ok(self):
        # pylint: disable=no-self-use
        watchdog_dict = {
            "node1": "",
            "node2": None,
            "node3": "/dev/watchdog",
        }
        assert_report_item_list_equal(
            cmd_sbd._validate_watchdog_dict(watchdog_dict),
            [
                (
                    Severities.ERROR,
                    reports.codes.WATCHDOG_INVALID,
                    {"watchdog": watchdog},
                )
                for watchdog in ["", None]
            ],
        )


class GetFullTargetDictTest(TestCase):
    def setUp(self):
        self.target_list = [
            RequestTarget("node{0}".format(i)) for i in range(1, 4)
        ]

    def test_not_using_default(self):
        target_dict = {"node" + str(i): "val" + str(i) for i in range(4)}
        expected = {
            self.target_list[0].label: "val1",
            self.target_list[1].label: "val2",
            self.target_list[2].label: "val3",
        }
        self.assertEqual(
            expected,
            cmd_sbd._get_full_target_dict(self.target_list, target_dict, None),
        )

    def test_using_default(self):
        target_dict = {"node" + str(i): "val" + str(i) for i in range(3)}
        default = "default"
        expected = {
            self.target_list[0].label: "val1",
            self.target_list[1].label: "val2",
            self.target_list[2].label: default,
        }
        self.assertEqual(
            expected,
            cmd_sbd._get_full_target_dict(
                self.target_list, target_dict, default
            ),
        )


class GetLocalSbdConfigTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_success(self):
        config_content = outdent(
            """\
            # This file has been generated by pcs.
            SBD_OPTS="-n node1"
            SBD_WATCHDOG_DEV=/dev/watchdog
            SBD_WATCHDOG_TIMEOUT=0
            """
        )
        self.config.fs.open(
            settings.sbd_config,
            mock.mock_open(read_data=config_content)(),
        )

        self.assertEqual(
            {
                "SBD_OPTS": '"-n node1"',
                "SBD_WATCHDOG_DEV": "/dev/watchdog",
                "SBD_WATCHDOG_TIMEOUT": "0",
            },
            cmd_sbd.get_local_sbd_config(self.env_assist.get_env()),
        )

    def test_file_error(self):
        node = "node"
        reason = "reason"
        self.config.fs.open(
            settings.sbd_config,
            side_effect=LibraryError(
                ReportItem.error(
                    reports.messages.UnableToGetSbdConfig(node, reason)
                )
            ),
        )
        assert_raise_library_error(
            lambda: cmd_sbd.get_local_sbd_config(self.env_assist.get_env()),
            (
                Severities.ERROR,
                reports.codes.UNABLE_TO_GET_SBD_CONFIG,
                {
                    "node": node,
                    "reason": reason,
                },
            ),
        )


class InitializeBlockDevicesTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    @staticmethod
    def fixture_invalid_value(option, value):
        return fixture.error(
            reports.codes.INVALID_OPTION_VALUE,
            option_name=option,
            option_value=value,
            allowed_values="a non-negative integer",
            cannot_be_empty=False,
            forbidden_characters=None,
        )

    def test_all_options(self):
        device_list = ["dev1", "dev2"]
        option_dict = {
            "watchdog-timeout": "1",
            "loop-timeout": "10",
            "allocate-timeout": "3",
            "msgwait-timeout": "2",
        }
        self.config.runner.sbd.initialize_devices(device_list, option_dict)
        cmd_sbd.initialize_block_devices(
            self.env_assist.get_env(), device_list, option_dict
        )
        self.env_assist.assert_reports(
            [
                (
                    Severities.INFO,
                    reports.codes.SBD_DEVICE_INITIALIZATION_STARTED,
                    {"device_list": device_list},
                ),
                (
                    Severities.INFO,
                    reports.codes.SBD_DEVICE_INITIALIZATION_SUCCESS,
                    {"device_list": device_list},
                ),
            ]
        )

    def test_no_options(self):
        device_list = ["dev1", "dev2"]
        self.config.runner.sbd.initialize_devices(device_list, {})
        cmd_sbd.initialize_block_devices(
            self.env_assist.get_env(), device_list, {}
        )
        self.env_assist.assert_reports(
            [
                (
                    Severities.INFO,
                    reports.codes.SBD_DEVICE_INITIALIZATION_STARTED,
                    {"device_list": device_list},
                ),
                (
                    Severities.INFO,
                    reports.codes.SBD_DEVICE_INITIALIZATION_SUCCESS,
                    {"device_list": device_list},
                ),
            ]
        )

    def test_validation_failed(self):
        option_dict = {
            "unknown_option": "val",
            "watchdog-timeout": "-1",
            "another_one": "-1",
            "loop-timeout": "-3",
            "allocate-timeout": "-3",
            "msgwait-timeout": "-2",
        }
        allowed_options = [
            "watchdog-timeout",
            "loop-timeout",
            "allocate-timeout",
            "msgwait-timeout",
        ]
        self.env_assist.assert_raise_library_error(
            lambda: cmd_sbd.initialize_block_devices(
                self.env_assist.get_env(), [], option_dict
            )
        )
        self.env_assist.assert_reports(
            [
                (
                    Severities.ERROR,
                    reports.codes.REQUIRED_OPTIONS_ARE_MISSING,
                    {
                        "option_names": ["device"],
                        "option_type": None,
                    },
                ),
                (
                    Severities.ERROR,
                    reports.codes.INVALID_OPTIONS,
                    {
                        "option_names": ["another_one", "unknown_option"],
                        "option_type": None,
                        "allowed": sorted(allowed_options),
                        "allowed_patterns": [],
                    },
                ),
            ]
            + [
                self.fixture_invalid_value(opt, option_dict[opt])
                for opt in allowed_options
            ],
        )


class GetLocalDevicesInfoTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_success(self):
        config_data = 'SBD_DEVICE="/dev1;/dev2"\n'
        self.config.services.is_enabled("sbd")
        self.config.fs.exists(settings.sbd_config, return_value=True)
        self.config.fs.open(
            settings.sbd_config, mock.mock_open(read_data=config_data)()
        )
        self.config.runner.sbd.get_device_info("/dev1", stdout="1")
        self.config.runner.sbd.get_device_info(
            "/dev2", stdout="2", name="list2"
        )
        expected_output = [
            {
                "device": "/dev1",
                "list": "1",
                "dump": None,
            },
            {
                "device": "/dev2",
                "list": "2",
                "dump": None,
            },
        ]
        self.assertEqual(
            expected_output,
            cmd_sbd.get_local_devices_info(self.env_assist.get_env()),
        )

    def test_with_dump(self):
        config_data = 'SBD_DEVICE="/dev1;/dev2"\n'
        self.config.services.is_enabled("sbd")
        self.config.fs.exists(settings.sbd_config, return_value=True)
        self.config.fs.open(
            settings.sbd_config, mock.mock_open(read_data=config_data)()
        )
        self.config.runner.sbd.get_device_info("/dev1", stdout="1")
        self.config.runner.sbd.get_device_dump("/dev1", stdout="2")
        self.config.runner.sbd.get_device_info(
            "/dev2", stdout="3", name="list2"
        )
        self.config.runner.sbd.get_device_dump(
            "/dev2", stdout="4", name="dump2"
        )
        expected_output = [
            {
                "device": "/dev1",
                "list": "1",
                "dump": "2",
            },
            {
                "device": "/dev2",
                "list": "3",
                "dump": "4",
            },
        ]
        self.assertEqual(
            expected_output,
            cmd_sbd.get_local_devices_info(
                self.env_assist.get_env(), dump=True
            ),
        )

    def test_no_config(self):
        self.config.services.is_enabled("sbd")
        self.config.fs.exists(settings.sbd_config, return_value=False)
        self.assertEqual(
            [], cmd_sbd.get_local_devices_info(self.env_assist.get_env())
        )

    def test_sbd_disabled(self):
        self.config.services.is_enabled("sbd", return_value=False)
        self.assertEqual(
            [], cmd_sbd.get_local_devices_info(self.env_assist.get_env())
        )

    def test_with_failures(self):
        config_data = 'SBD_DEVICE="/dev1;/dev2;/dev3"\n'
        self.config.services.is_enabled("sbd")
        self.config.fs.exists(settings.sbd_config, return_value=True)
        self.config.fs.open(
            settings.sbd_config, mock.mock_open(read_data=config_data)()
        )
        self.config.runner.sbd.get_device_info(
            "/dev1", stdout="1", return_code=1
        )
        self.config.runner.sbd.get_device_info(
            "/dev2", stdout="2", name="list2"
        )
        self.config.runner.sbd.get_device_dump(
            "/dev2", stdout="4", return_code=1
        )
        self.config.runner.sbd.get_device_info(
            "/dev3", stdout="5", name="list3"
        )
        self.config.runner.sbd.get_device_dump(
            "/dev3", stdout="6", name="dump2"
        )
        expected_output = [
            {
                "device": "/dev1",
                "list": None,
                "dump": None,
            },
            {
                "device": "/dev2",
                "list": "2",
                "dump": None,
            },
            {
                "device": "/dev3",
                "list": "5",
                "dump": "6",
            },
        ]
        self.assertEqual(
            expected_output,
            cmd_sbd.get_local_devices_info(
                self.env_assist.get_env(), dump=True
            ),
        )
        self.env_assist.assert_reports(
            [
                (
                    Severities.WARNING,
                    reports.codes.SBD_DEVICE_LIST_ERROR,
                    {"device": "/dev1", "reason": "1"},
                ),
                (
                    Severities.WARNING,
                    reports.codes.SBD_DEVICE_DUMP_ERROR,
                    {"device": "/dev2", "reason": "4"},
                ),
            ]
        )


class SetMessageTest(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_empty_options(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_sbd.set_message(self.env_assist.get_env(), "", "", "")
        )
        self.env_assist.assert_reports(
            [
                (
                    Severities.ERROR,
                    reports.codes.REQUIRED_OPTIONS_ARE_MISSING,
                    {
                        "option_names": ["device", "node"],
                        "option_type": None,
                    },
                ),
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="message",
                    option_value="",
                    allowed_values=settings.sbd_message_types,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_invalid_message_type(self):
        self.env_assist.assert_raise_library_error(
            lambda: cmd_sbd.set_message(
                self.env_assist.get_env(), "device", "node1", "message"
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="message",
                    option_value="message",
                    allowed_values=settings.sbd_message_types,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                )
            ],
        )

    def test_success(self):
        self.config.runner.sbd.set_device_message("device", "node", "test")
        cmd_sbd.set_message(self.env_assist.get_env(), "device", "node", "test")

    def test_failure(self):
        self.config.runner.sbd.set_device_message(
            "device", "node", "test", stderr="error", return_code=1
        )
        self.env_assist.assert_raise_library_error(
            lambda: cmd_sbd.set_message(
                self.env_assist.get_env(), "device", "node", "test"
            ),
            [
                (
                    Severities.ERROR,
                    reports.codes.SBD_DEVICE_MESSAGE_ERROR,
                    {
                        "device": "device",
                        "node": "node",
                        "sbd_message": "test",
                        "reason": "error",
                    },
                )
            ],
            expected_in_processor=False,
        )



from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs.common import const
from pcs.common.pacemaker.resource.operations import CibResourceOperationDto
from pcs.common.reports import ReportItemSeverity as severities
from pcs.common.reports import codes as report_codes
from pcs.lib.cib.resource import operations
from pcs.lib.validate import ValuePair

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal
from pcs_test.tools.misc import create_patcher

# pylint: disable=no-self-use


def _operation_fixture(name, interval="", role=None):
    return CibResourceOperationDto(
        id="",
        name=name,
        interval=interval,
        description=None,
        start_delay=None,
        interval_origin=None,
        timeout=None,
        enabled=None,
        record_pending=None,
        role=role,
        on_fail=None,
        meta_attributes=[],
        instance_attributes=[],
    )


patch_operations = create_patcher("pcs.lib.cib.resource.operations")


@patch_operations("_get_remaining_defaults")
@patch_operations("complete_operations_options")
@patch_operations("validate_different_intervals")
@patch_operations("_validate_operation_list")
@patch_operations("_normalized_to_operations")
@patch_operations("_operations_to_normalized")
class Prepare(TestCase):
    def test_prepare(
        self,
        operations_to_normalized,
        normalized_to_operations,
        validate_operation_list,
        validate_different_intervals,
        complete_operations_options,
        get_remaining_defaults,
    ):
        new_role_names_supported = False
        validate_operation_list.return_value = ["options_report"]
        validate_different_intervals.return_value = [
            "different_interval_report"
        ]
        operations_to_normalized.return_value = [
            {"name": ValuePair("Start", "start")},
            {"name": ValuePair("Monitor", "monitor")},
        ]
        normalized_to_operations.return_value = [
            {"name": "start"},
            {"name": "monitor"},
        ]

        report_processor = mock.MagicMock()
        report_processor.report_list.return_value = report_processor
        report_processor.has_errors = False

        raw_operation_list = [
            {"name": "Start"},
            {"name": "Monitor"},
        ]
        default_operation_list = [{"name": "stop", "interval": "10s"}]
        completed_default_operation_list = [
            {"name": "start", "interval": "0s"},
            {"name": "monitor", "interval": "60s"},
        ]
        allowed_operation_name_list = ["start", "stop", "monitor"]
        allow_invalid = True

        complete_operations_options.return_value = (
            completed_default_operation_list
        )

        operations.prepare(
            report_processor,
            raw_operation_list,
            default_operation_list,
            allowed_operation_name_list,
            new_role_names_supported,
            allow_invalid,
        )

        operations_to_normalized.assert_called_once_with(raw_operation_list)
        normalized_to_operations.assert_called_once_with(
            operations_to_normalized.return_value, new_role_names_supported
        )
        validate_operation_list.assert_called_once_with(
            operations_to_normalized.return_value,
            allowed_operation_name_list,
            allow_invalid,
        )
        validate_different_intervals.assert_called_once_with(
            normalized_to_operations.return_value
        )
        complete_operations_options.assert_has_calls(
            [mock.call(normalized_to_operations.return_value)]
        )
        get_remaining_defaults.assert_called_once_with(
            normalized_to_operations.return_value,
            default_operation_list,
        )
        report_processor.report_list.assert_has_calls(
            [
                mock.call(["options_report", "different_interval_report"]),
                mock.call([]),
            ]
        )


class ValidateDifferentIntervals(TestCase):
    def test_return_empty_reports_on_empty_list(self):
        operations.validate_different_intervals([])

    def test_return_empty_reports_on_operations_without_duplication(self):
        operations.validate_different_intervals(
            [
                {"name": "monitor", "interval": "10s"},
                {"name": "monitor", "interval": "5s"},
                {"name": "start", "interval": "5s"},
            ]
        )

    def test_return_report_on_duplicated_intervals(self):
        assert_report_item_list_equal(
            operations.validate_different_intervals(
                [
                    {"name": "monitor", "interval": "3600s"},
                    {"name": "monitor", "interval": "60m"},
                    {"name": "monitor", "interval": "1h"},
                    {"name": "monitor", "interval": "60s"},
                    {"name": "monitor", "interval": "1m"},
                    {"name": "monitor", "interval": "5s"},
                ]
            ),
            [
                (
                    severities.ERROR,
                    report_codes.RESOURCE_OPERATION_INTERVAL_DUPLICATION,
                    {
                        "duplications": {
                            "monitor": [
                                ["3600s", "60m", "1h"],
                                ["60s", "1m"],
                            ],
                        },
                    },
                )
            ],
        )


class MakeUniqueIntervals(TestCase):
    def test_return_copy_input_when_no_interval_duplication(self):
        operation_list = [
            _operation_fixture("monitor", "10s"),
            _operation_fixture("monitor", "5s"),
            _operation_fixture("monitor"),
            _operation_fixture("start", "5s"),
        ]
        (
            report_list,
            new_operation_list,
        ) = operations.uniquify_operations_intervals(operation_list)
        assert_report_item_list_equal(report_list, [])
        self.assertEqual(operation_list, new_operation_list)

    def test_adopt_duplicit_values(self):
        (
            report_list,
            new_operation_list,
        ) = operations.uniquify_operations_intervals(
            [
                _operation_fixture("monitor", "60s"),
                _operation_fixture("monitor", "1m"),
                _operation_fixture("monitor", "5s"),
                _operation_fixture("monitor", "6s"),
                _operation_fixture("monitor", "5s"),
                _operation_fixture("start", "5s"),
            ]
        )
        self.assertEqual(
            new_operation_list,
            [
                _operation_fixture("monitor", "60s"),
                _operation_fixture("monitor", "61"),
                _operation_fixture("monitor", "5s"),
                _operation_fixture("monitor", "6s"),
                _operation_fixture("monitor", "7"),
                _operation_fixture("start", "5s"),
            ],
        )

        assert_report_item_list_equal(
            report_list,
            [
                (
                    severities.WARNING,
                    report_codes.RESOURCE_OPERATION_INTERVAL_ADAPTED,
                    {
                        "operation_name": "monitor",
                        "original_interval": "1m",
                        "adapted_interval": "61",
                    },
                ),
                (
                    severities.WARNING,
                    report_codes.RESOURCE_OPERATION_INTERVAL_ADAPTED,
                    {
                        "operation_name": "monitor",
                        "original_interval": "5s",
                        "adapted_interval": "7",
                    },
                ),
            ],
        )

    def test_keep_duplicit_values_when_are_not_valid_interval(self):
        operation_list = [
            _operation_fixture("monitor", "some"),
            _operation_fixture("monitor", "some"),
        ]
        (
            report_list,
            new_operation_list,
        ) = operations.uniquify_operations_intervals(operation_list)
        assert_report_item_list_equal(report_list, [])
        self.assertEqual(new_operation_list, operation_list)


class Normalize(TestCase):
    # pylint: disable=protected-access
    def test_return_operation_with_the_same_values(self):
        operation = {
            "name": "monitor",
            "role": const.PCMK_ROLE_PROMOTED_LEGACY,
            "timeout": "10",
        }

        self.assertEqual(
            operation,
            {
                key: operations._normalize(key, value)
                for key, value in operation.items()
            },
        )

    def test_return_operation_with_normalized_values(self):
        self.assertEqual(
            {
                "name": "monitor",
                "role": const.PCMK_ROLE_PROMOTED_LEGACY,
                "timeout": "10",
                "on-fail": "ignore",
                "record-pending": "true",
                "enabled": "1",
            },
            {
                key: operations._normalize(key, value)
                for key, value in {
                    "name": "monitor",
                    "role": "master",
                    "timeout": "10",
                    "on-fail": "Ignore",
                    "record-pending": "True",
                    "enabled": "1",
                }.items()
            },
        )


class ValidateOperation(TestCase):
    def assert_operation_produces_report(self, operation, report_list):
        # pylint: disable=protected-access
        assert_report_item_list_equal(
            operations._validate_operation_list(
                [operation],
                ["monitor"],
            ),
            report_list,
        )

    def test_return_empty_report_on_valid_operation(self):
        self.assert_operation_produces_report(
            {"name": "monitor", "role": "Master"}, []
        )

    def test_validate_all_individual_options(self):
        self.assert_operation_produces_report(
            {
                "name": "monitro",
                "role": "a",
                "on-fail": "b",
                "record-pending": "c",
                "enabled": "d",
                "id": "0",
                "unknown": "invalid",
            },
            [
                fixture.error(
                    report_codes.INVALID_OPTIONS,
                    option_names=["unknown"],
                    option_type="resource operation",
                    allowed=[
                        "OCF_CHECK_LEVEL",
                        "description",
                        "enabled",
                        "id",
                        "interval",
                        "interval-origin",
                        "name",
                        "on-fail",
                        "record-pending",
                        "role",
                        "start-delay",
                        "timeout",
                    ],
                    allowed_patterns=[],
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    force_code=report_codes.FORCE,
                    option_value="monitro",
                    option_name="operation name",
                    allowed_values=["monitor"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="a",
                    option_name="role",
                    allowed_values=const.PCMK_ROLES,
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="b",
                    option_name="on-fail",
                    allowed_values=tuple(
                        sorted(
                            [
                                "block",
                                "demote",
                                "fence",
                                "ignore",
                                "restart",
                                "restart-container",
                                "standby",
                                "stop",
                            ]
                        )
                    ),
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="c",
                    option_name="record-pending",
                    allowed_values=["0", "1", "true", "false"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_OPTION_VALUE,
                    option_value="d",
                    option_name="enabled",
                    allowed_values=["0", "1", "true", "false"],
                    cannot_be_empty=False,
                    forbidden_characters=None,
                ),
                fixture.error(
                    report_codes.INVALID_ID_BAD_CHAR,
                    id="0",
                    id_description="operation id",
                    is_first_char=True,
                    invalid_character="0",
                ),
            ],
        )

    def test_return_error_when_unknown_operation_attribute(self):
        self.assert_operation_produces_report(
            {
                "name": "monitor",
                "unknown": "invalid",
            },
            [
                (
                    severities.ERROR,
                    report_codes.INVALID_OPTIONS,
                    {
                        "option_names": ["unknown"],
                        "option_type": "resource operation",
                        "allowed": sorted(operations.ATTRIBUTES),
                        "allowed_patterns": [],
                    },
                    None,
                ),
            ],
        )

    def test_return_error_when_missing_key_name(self):
        self.assert_operation_produces_report(
            {"role": "Master"},
            [
                (
                    severities.ERROR,
                    report_codes.REQUIRED_OPTIONS_ARE_MISSING,
                    {
                        "option_names": ["name"],
                        "option_type": "resource operation",
                    },
                    None,
                ),
            ],
        )

    def test_return_error_when_both_interval_origin_and_start_delay(self):
        self.assert_operation_produces_report(
            {
                "name": "monitor",
                "interval-origin": "a",
                "start-delay": "b",
            },
            [
                (
                    severities.ERROR,
                    report_codes.MUTUALLY_EXCLUSIVE_OPTIONS,
                    {
                        "option_names": ["interval-origin", "start-delay"],
                        "option_type": "resource operation",
                    },
                    None,
                ),
            ],
        )

    def test_return_error_on_invalid_id(self):
        self.assert_operation_produces_report(
            {
                "name": "monitor",
                "id": "a#b",
            },
            [
                (
                    severities.ERROR,
                    report_codes.INVALID_ID_BAD_CHAR,
                    {
                        "id": "a#b",
                        "id_description": "operation id",
                        "invalid_character": "#",
                        "is_first_char": False,
                    },
                    None,
                ),
            ],
        )


class GetRemainingDefaults(TestCase):
    def test_success(self):
        # pylint: disable=protected-access
        self.assertEqual(
            operations._get_remaining_defaults(
                operation_list=[{"name": "monitor"}],
                default_operation_list=[
                    _operation_fixture("monitor"),
                    _operation_fixture("start"),
                    _operation_fixture("stop"),
                ],
            ),
            [
                _operation_fixture("start"),
                _operation_fixture("stop"),
            ],
        )


class GetResourceOperations(TestCase):
    resource_el = etree.fromstring(
        """
        <primitive class="ocf" id="dummy" provider="pacemaker" type="Stateful">
            <operations>
                <op id="dummy-start" interval="0s" name="start" timeout="20"/>
                <op id="dummy-stop" interval="0s" name="stop" timeout="20"/>
                <op id="dummy-monitor-m" interval="10" name="monitor"
                    role="Master" timeout="20"/>
                <op id="dummy-monitor-s" interval="11" name="monitor"
                    role="Slave" timeout="20"/>
            </operations>
        </primitive>
    """
    )
    resource_noop_el = etree.fromstring(
        """
        <primitive class="ocf" id="dummy" provider="pacemaker" type="Stateful">
        </primitive>
    """
    )

    def assert_op_list(self, op_list, expected_ids):
        self.assertEqual([op.attrib.get("id") for op in op_list], expected_ids)

    def test_all_operations(self):
        self.assert_op_list(
            operations.get_resource_operations(self.resource_el),
            ["dummy-start", "dummy-stop", "dummy-monitor-m", "dummy-monitor-s"],
        )

    def test_filter_operations(self):
        self.assert_op_list(
            operations.get_resource_operations(self.resource_el, ["start"]),
            ["dummy-start"],
        )

    def test_filter_more_operations(self):
        self.assert_op_list(
            operations.get_resource_operations(
                self.resource_el, ["monitor", "stop"]
            ),
            ["dummy-stop", "dummy-monitor-m", "dummy-monitor-s"],
        )

    def test_filter_none(self):
        self.assert_op_list(
            operations.get_resource_operations(self.resource_el, ["promote"]),
            [],
        )

    def test_no_operations(self):
        self.assert_op_list(
            operations.get_resource_operations(self.resource_noop_el), []
        )



from pcs.common import reports
from pcs.common.reports.item import ReportItem
from pcs.lib import validate
from pcs.lib.cib.node import PacemakerNode
from pcs.lib.cib.nvpair import (
    arrange_first_meta_attributes,
    get_meta_attribute_value,
    has_meta_attribute,
)
from pcs.lib.cib.tools import does_id_exist

# TODO pcs currently does not care about multiple meta_attributes and here
# we don't care as well
GUEST_OPTIONS = [
    "remote-port",
    "remote-addr",
    "remote-connect-timeout",
]


def validate_conflicts(
    tree, existing_nodes_names, existing_nodes_addrs, node_name, options
):
    report_list = []
    if (
        does_id_exist(tree, node_name)
        or node_name in existing_nodes_names
        or ("remote-addr" not in options and node_name in existing_nodes_addrs)
    ):
        report_list.append(
            ReportItem.error(reports.messages.IdAlreadyExists(node_name))
        )

    if (
        "remote-addr" in options
        and options["remote-addr"] in existing_nodes_addrs
    ):
        report_list.append(
            ReportItem.error(
                reports.messages.IdAlreadyExists(options["remote-addr"])
            )
        )
    return report_list


def is_node_name_in_options(options):
    return "remote-node" in options


def get_guest_option_value(options, default=None):
    """
    Commandline options: no options
    """
    return options.get("remote-node", default)


def validate_set_as_guest(
    tree, existing_nodes_names, existing_nodes_addrs, node_name, options
):
    validator_list = [
        validate.NamesIn(GUEST_OPTIONS, option_type="guest"),
        validate.ValueTimeInterval("remote-connect-timeout"),
        validate.ValuePortNumber("remote-port"),
    ]
    return (
        validate.ValidatorAll(validator_list).validate(options)
        + validate.ValueNotEmpty("node name", None).validate(
            {"node name": node_name.strip()}
        )
        + validate_conflicts(
            tree, existing_nodes_names, existing_nodes_addrs, node_name, options
        )
    )


def is_guest_node(resource_element):
    """
    Return True if resource_element is already set as guest node.

    etree.Element resource_element is a search element
    """
    return has_meta_attribute(resource_element, "remote-node")


def validate_is_not_guest(resource_element):
    """
    etree.Element resource_element
    """
    if not is_guest_node(resource_element):
        return []

    return [
        ReportItem.error(
            reports.messages.ResourceIsGuestNodeAlready(
                resource_element.attrib["id"]
            )
        )
    ]


def set_as_guest(
    resource_element,
    id_provider,
    node,
    addr=None,
    port=None,
    connect_timeout=None,
):
    """
    Set resource as guest node.

    etree.Element resource_element

    """
    meta_options = {"remote-node": str(node)}
    if addr:
        meta_options["remote-addr"] = str(addr)
    if port:
        meta_options["remote-port"] = str(port)
    if connect_timeout:
        meta_options["remote-connect-timeout"] = str(connect_timeout)

    arrange_first_meta_attributes(resource_element, meta_options, id_provider)


def unset_guest(resource_element):
    """
    Unset resource as guest node.

    etree.Element resource_element
    """
    # Do not ever remove the nvset element, even if it is empty. There may be
    # ACLs set in pacemaker which allow "write" for nvpairs (adding, changing
    # and removing) but not nvsets. In such a case, removing the nvset would
    # cause the whole change to be rejected by pacemaker with a "permission
    # denied" message.
    # https://bugzilla.redhat.com/show_bug.cgi?id=1642514
    guest_nvpair_list = resource_element.xpath(
        "./meta_attributes/nvpair[{0}]".format(
            " or ".join(
                [
                    '@name="{0}"'.format(option)
                    for option in (GUEST_OPTIONS + ["remote-node"])
                ]
            )
        )
    )
    for nvpair in guest_nvpair_list:
        meta_attributes = nvpair.getparent()
        meta_attributes.remove(nvpair)


def get_node_name_from_options(meta_options, default=None):
    """
    Return node_name from meta options.
    dict meta_options
    """
    return meta_options.get("remote-node", default)


def get_node_name_from_resource(resource_element):
    """
    Return the node name from a remote node resource, None for other resources

    etree.Element resource_element
    """
    return get_meta_attribute_value(resource_element, "remote-node")


def find_node_list(tree):
    """
    Return list of guest nodes from the specified element tree

    etree.Element tree -- an element to search guest nodes in
    """
    node_list = []
    for meta_attrs in tree.xpath(
        """
            .//primitive
                /meta_attributes[
                    nvpair[
                        @name="remote-node"
                        and
                        string-length(@value) > 0
                    ]
                ]
        """
    ):
        host = None
        name = None
        for nvpair in meta_attrs:
            if nvpair.attrib.get("name", "") == "remote-addr":
                host = nvpair.attrib["value"]
            if nvpair.attrib.get("name", "") == "remote-node":
                name = nvpair.attrib["value"]
                if host is None:
                    host = name
        if name:
            # The name is never empty as we only loop through elements with
            # non-empty names. It's just we loop through nvpairs instead of
            # reading the name directly.
            node_list.append(PacemakerNode(name, host))
    return node_list


def find_node_resources(resources_section, node_identifier):
    """
    Return list of etree.Element primitives that are guest nodes.

    etree.Element resources_section is a researched element
    string node_identifier could be id of resource, node name or node address
    """
    resources = resources_section.xpath(
        """
        .//primitive[
            (
                @id=$node_id
                and
                meta_attributes[
                    nvpair[
                        @name="remote-node"
                        and
                        string-length(@value) > 0
                    ]
                ]
            )
            or
            meta_attributes[
                nvpair[
                    @name="remote-node"
                    and
                    string-length(@value) > 0
                ]
                and
                nvpair[
                    (
                        @name="remote-addr"
                        or
                        @name="remote-node"
                    )
                    and
                    @value=$node_id
                ]
            ]
        ]
    """,
        node_id=node_identifier,
    )
    return resources



from typing import (
    Dict,
    Mapping,
    Optional,
)

# TODO temporary definition, we want to overhaul the whole 'resource create'
# command using this and implement a proper dataclass for resource operations
ResourceOperationIn = Mapping[str, Optional[str]]
ResourceOperationOut = Dict[str, Optional[str]]
ResourceOperationFilteredIn = Mapping[str, str]
ResourceOperationFilteredOut = Dict[str, str]



# coding=utf-8
from unittest import TestCase

from pcs.common import const
from pcs.common.interface.dto import from_dict
from pcs.common.pacemaker.resource.operations import (
    CibResourceOperationDto,
    ListCibResourceOperationDto,
)
from pcs.common.reports import codes as report_codes
from pcs.common.resource_agent.dto import (
    ListResourceAgentNameDto,
    ResourceAgentActionDto,
    ResourceAgentMetadataDto,
    ResourceAgentNameDto,
    ResourceAgentParameterDto,
)
from pcs.lib.commands import resource_agent as lib
from pcs.lib.resource_agent import ResourceAgentName

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools


def _operation_fixture(name, interval="", role=None, timeout=None):
    return CibResourceOperationDto(
        id="",
        name=name,
        interval=interval,
        description=None,
        start_delay=None,
        interval_origin=None,
        timeout=timeout,
        enabled=None,
        record_pending=None,
        role=role,
        on_fail=None,
        meta_attributes=[],
        instance_attributes=[],
    )


class ListStandards(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_success(self):
        standards = ["service", "lsb", "ocf", "systemd", "lsb", "nagios"]
        self.config.runner.pcmk.list_agents_standards("\n".join(standards))
        self.assertEqual(
            lib.list_standards(self.env_assist.get_env()),
            sorted(set(standards)),
        )


class ListOcfProviders(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_success(self):
        providers = ["pacemaker", "booth", "openstack", "booth", "heartbeat"]
        self.config.runner.pcmk.list_agents_ocf_providers("\n".join(providers))
        self.assertEqual(
            lib.list_ocf_providers(self.env_assist.get_env()),
            sorted(set(providers)),
        )


class ListAgentsForStandardAndProvider(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def _assert_standard_provider_specified(self, standard):
        agents = ["Delay", "Stateful", "Delay", "Dummy"]
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:pacemaker", "\n".join(agents)
        )
        self.assertEqual(
            lib.list_agents_for_standard_and_provider(
                self.env_assist.get_env(), standard
            ),
            sorted(set(agents)),
        )

    def test_standard_provider_specified_1(self):
        self._assert_standard_provider_specified("ocf:pacemaker")

    def test_standard_provider_specified_2(self):
        self._assert_standard_provider_specified("ocf:pacemaker:")

    def test_standard_provider_not_specified(self):
        agents_pacemaker = ["Delay", "Dummy", "Stateful"]
        agents_heartbeat = ["Delay"]
        agents_service = ["corosync", "pacemaker", "pcsd"]
        self.config.runner.pcmk.list_agents_standards(
            "\n".join(["service", "stonith", "ocf"])
        )
        self.config.runner.pcmk.list_agents_ocf_providers(
            "\n".join(["heartbeat", "pacemaker"])
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:heartbeat",
            "\n".join(agents_heartbeat),
            name="runner.pcmk.list_agents_ocf_providers.heartbeat",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:pacemaker",
            "\n".join(agents_pacemaker),
            name="runner.pcmk.list_agents_ocf_providers.pacemaker",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "service",
            "\n".join(agents_service),
            name="runner.pcmk.list_agents_ocf_providers.service",
        )
        self.assertEqual(
            lib.list_agents_for_standard_and_provider(
                self.env_assist.get_env()
            ),
            sorted(
                agents_pacemaker + agents_heartbeat + agents_service,
                key=str.lower,
            ),
        )


class ListAgents(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.env_assist, self.config = get_env_tools(test_case=self)

        self.config.runner.pcmk.list_agents_standards(
            "\n".join(["service", "ocf"])
        )
        self.config.runner.pcmk.list_agents_ocf_providers("\n".join(["test"]))
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:test",
            "\n".join(["Stateful", "Delay"]),
            name="runner.pcmk.list_agents_ocf_providers.ocf_test",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "service",
            "\n".join(["corosync", "pacemaker_remote"]),
            name="runner.pcmk.list_agents_ocf_providers.service",
        )

    @staticmethod
    def _fixture_agent_struct(name):
        return {
            "name": name.full_name,
            "standard": name.standard,
            "provider": name.provider,
            "type": name.type,
            "shortdesc": None,
            "longdesc": None,
            "parameters": [],
            "actions": [],
            "default_actions": [],
        }

    @staticmethod
    def _fixture_agent_metadata(name):
        return f"""
            <resource-agent name="{name}">
                <shortdesc>short {name}</shortdesc>
                <longdesc>long {name}</longdesc>
                <parameters>
                </parameters>
                <actions>
                </actions>
            </resource-agent>
            """

    def test_list_all(self):
        self.assertEqual(
            lib.list_agents(self.env_assist.get_env(), False, None),
            [
                self._fixture_agent_struct(
                    ResourceAgentName("ocf", "test", "Delay")
                ),
                self._fixture_agent_struct(
                    ResourceAgentName("ocf", "test", "Stateful")
                ),
                self._fixture_agent_struct(
                    ResourceAgentName("service", None, "corosync")
                ),
                self._fixture_agent_struct(
                    ResourceAgentName("service", None, "pacemaker_remote")
                ),
            ],
        )

    def test_search(self):
        self.assertEqual(
            lib.list_agents(self.env_assist.get_env(), False, "te"),
            [
                self._fixture_agent_struct(
                    ResourceAgentName("ocf", "test", "Delay")
                ),
                self._fixture_agent_struct(
                    ResourceAgentName("ocf", "test", "Stateful")
                ),
                self._fixture_agent_struct(
                    ResourceAgentName("service", None, "pacemaker_remote")
                ),
            ],
        )

    def test_describe(self):
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:test:Delay",
            stdout=self._fixture_agent_metadata("ocf:test:Delay"),
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
            name="runner.pcmk.load_agent.delay",
        )
        self.config.runner.pcmk.load_agent(
            agent_name="ocf:test:Stateful",
            agent_is_missing=True,
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
            name="runner.pcmk.load_agent.stateful",
        )
        self.config.runner.pcmk.load_agent(
            agent_name="service:corosync",
            stdout=self._fixture_agent_metadata("service:corosync"),
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
            name="runner.pcmk.load_agent.corosync",
        )
        self.config.runner.pcmk.load_agent(
            agent_name="service:pacemaker_remote",
            stdout=self._fixture_agent_metadata("service:pacemaker_remote"),
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
            name="runner.pcmk.load_agent.pacemaker_remote",
        )

        agent_stub = {
            "parameters": [],
            "actions": [],
            "default_actions": [
                {
                    "interval": "60s",
                    "name": "monitor",
                    "OCF_CHECK_LEVEL": None,
                    "automatic": False,
                    "on_target": False,
                    "role": None,
                    "start-delay": None,
                    "timeout": None,
                }
            ],
        }
        self.assertEqual(
            lib.list_agents(self.env_assist.get_env(), True, None),
            [
                dict(
                    name="ocf:test:Delay",
                    standard="ocf",
                    provider="test",
                    type="Delay",
                    shortdesc="short ocf:test:Delay",
                    longdesc="long ocf:test:Delay",
                    **agent_stub,
                ),
                dict(
                    name="service:corosync",
                    standard="service",
                    provider=None,
                    type="corosync",
                    shortdesc="short service:corosync",
                    longdesc="long service:corosync",
                    **agent_stub,
                ),
                dict(
                    name="service:pacemaker_remote",
                    standard="service",
                    provider=None,
                    type="pacemaker_remote",
                    shortdesc="short service:pacemaker_remote",
                    longdesc="long service:pacemaker_remote",
                    **agent_stub,
                ),
            ],
        )


class ActionToOperation(TestCase):
    # pylint: disable=protected-access
    @staticmethod
    def _action_dict(action):
        all_keys = {
            "name": "",
            "timeout": None,
            "interval": None,
            "role": None,
            "start-delay": None,
            "OCF_CHECK_LEVEL": None,
            "automatic": False,
            "on_target": False,
        }
        all_keys.update(action)
        return all_keys

    @staticmethod
    def _action_dto(action):
        all_keys = {
            "name": "",
            "timeout": None,
            "interval": None,
            "role": None,
            "start-delay": None,
            "depth": None,
            "automatic": False,
            "on_target": False,
        }
        all_keys.update(action)
        return from_dict(ResourceAgentActionDto, all_keys)

    def test_remove_depth_with_0(self):
        self.assertEqual(
            lib._action_to_operation(
                self._action_dto(
                    {"name": "monitor", "timeout": "20", "depth": "0"},
                )
            ),
            self._action_dict({"name": "monitor", "timeout": "20"}),
        )

    def test_transform_depth_to_ocf_check_level(self):
        self.assertEqual(
            lib._action_to_operation(
                self._action_dto(
                    {"name": "monitor", "timeout": "20", "depth": "1"},
                )
            ),
            self._action_dict(
                {"name": "monitor", "timeout": "20", "OCF_CHECK_LEVEL": "1"}
            ),
        )


class DescribeAgent(TestCase):
    def setUp(self):
        self.maxDiff = None
        self.env_assist, self.config = get_env_tools(test_case=self)

        self.trace_parameters = [
            {
                "advanced": True,
                "default": "0",
                "deprecated": False,
                "deprecated_by": [],
                "deprecated_desc": None,
                "enum_values": None,
                "longdesc": "Set to 1 to turn on resource agent tracing"
                " (expect large output) The trace output will be "
                "saved to trace_file, if set, or by default to "
                "$HA_VARRUN/ra_trace/<type>/<id>.<action>."
                "<timestamp> e.g. $HA_VARRUN/ra_trace/oracle/db."
                "start.2012-11-27.08:37:08",
                "name": "trace_ra",
                "required": False,
                "shortdesc": "Set to 1 to turn on resource agent "
                "tracing (expect large output)",
                "type": "integer",
                "unique_group": None,
                "reloadable": False,
            },
            {
                "advanced": True,
                "default": "",
                "deprecated": False,
                "deprecated_by": [],
                "deprecated_desc": None,
                "enum_values": None,
                "longdesc": (
                    "Path to a file to store resource agent tracing log"
                ),
                "name": "trace_file",
                "required": False,
                "shortdesc": (
                    "Path to a file to store resource agent tracing log"
                ),
                "type": "string",
                "unique_group": None,
                "reloadable": False,
            },
        ]

    def test_full_name_and_utf8_success(self):
        full_name = "ocf:heartbeat:Dummy"
        self.config.runner.pcmk.load_agent(
            agent_filename="resource_agent_ocf_heartbeat_dummy_utf8.xml",
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
        )
        self.assertEqual(
            lib.describe_agent(self.env_assist.get_env(), full_name),
            {
                "name": full_name,
                "standard": "ocf",
                "provider": "heartbeat",
                "type": "Dummy",
                "shortdesc": "Example stateless resource agent: ®",
                "longdesc": "This is a Dummy Resource Agent for testing utf-8"
                " in metadata: ®",
                "parameters": [
                    {
                        "advanced": False,
                        "default": "/var/run/resource-agents/Dummy-®.state",
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "enum_values": None,
                        "longdesc": (
                            "Location to store the resource state in: ®"
                        ),
                        "name": "state-®",
                        "required": False,
                        "shortdesc": "State file: ®",
                        "type": "string",
                        "unique_group": "_pcs_unique_group_state-®",
                        "reloadable": True,
                    },
                ]
                + self.trace_parameters,
                "actions": [
                    {
                        "name": "start",
                        "timeout": "20",
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                    },
                    {
                        "name": "stop",
                        "timeout": "20",
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                    },
                    {
                        "name": "monitor",
                        "interval": "10",
                        "timeout": "20",
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                    },
                    {
                        "name": "meta-data",
                        "timeout": "5",
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                    },
                    {
                        "name": "validate-all",
                        "timeout": "20",
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                    },
                    {
                        "name": "custom-®",
                        "timeout": "20",
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                    },
                ],
                "default_actions": [
                    {
                        "name": "start",
                        "interval": "0s",
                        "timeout": "20",
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                    },
                    {
                        "name": "stop",
                        "interval": "0s",
                        "timeout": "20",
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                    },
                    {
                        "name": "monitor",
                        "interval": "10",
                        "timeout": "20",
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                    },
                    {
                        "name": "custom-®",
                        "interval": "0s",
                        "timeout": "20",
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                    },
                ],
            },
        )

    def test_guess_success(self):
        self.config.runner.pcmk.list_agents_standards(
            "\n".join(["service", "ocf"])
        )
        self.config.runner.pcmk.list_agents_ocf_providers(
            "\n".join(["heartbeat", "pacemaker"])
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:heartbeat",
            "\n".join(["agent1", "Dummy", "agent2"]),
            name="runner.pcmk.list_agents_ocf_providers.heartbeat",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:pacemaker",
            "\n".join(["agent1"]),
            name="runner.pcmk.list_agents_ocf_providers.pacemaker",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "service",
            "\n".join(["agent1"]),
            name="runner.pcmk.list_agents_ocf_providers.service",
        )
        self.config.runner.pcmk.load_agent(
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
        )
        self.assertEqual(
            lib.describe_agent(self.env_assist.get_env(), "dummy"),
            {
                "name": "ocf:heartbeat:Dummy",
                "standard": "ocf",
                "provider": "heartbeat",
                "type": "Dummy",
                "shortdesc": "Example stateless resource agent",
                "longdesc": (
                    "This is a Dummy Resource Agent. It does absolutely nothing "
                    "except \nkeep track of whether its running or not.\nIts "
                    "purpose in life is for testing and to serve as a template "
                    "for RA writers.\n\nNB: Please pay attention to the timeouts "
                    "specified in the actions\nsection below. They should be "
                    "meaningful for the kind of resource\nthe agent manages. "
                    "They should be the minimum advised timeouts,\nbut they "
                    "shouldn't/cannot cover _all_ possible resource\ninstances. "
                    "So, try to be neither overly generous nor too stingy,\nbut "
                    "moderate. The minimum timeouts should never be below 10 seconds."
                ),
                "parameters": [
                    {
                        "name": "state",
                        "shortdesc": "State file",
                        "longdesc": "Location to store the resource state in.",
                        "type": "string",
                        "default": "/var/run/resource-agents/Dummy-undef.state",
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": "_pcs_unique_group_state",
                        "reloadable": True,
                    },
                    {
                        "name": "fake",
                        "shortdesc": (
                            "Fake attribute that can be changed to cause a reload"
                        ),
                        "longdesc": (
                            "Fake attribute that can be changed to cause a reload"
                        ),
                        "type": "string",
                        "default": "dummy",
                        "enum_values": None,
                        "required": False,
                        "advanced": False,
                        "deprecated": False,
                        "deprecated_by": [],
                        "deprecated_desc": None,
                        "unique_group": None,
                        "reloadable": False,
                    },
                ]
                + self.trace_parameters,
                "actions": [
                    {
                        "name": "start",
                        "timeout": "20",
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                    },
                    {
                        "name": "stop",
                        "timeout": "20",
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                    },
                    {
                        "name": "monitor",
                        "timeout": "20",
                        "interval": "10",
                        "automatic": False,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                    },
                    {
                        "name": "reload",
                        "timeout": "20",
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                    },
                    {
                        "name": "migrate_to",
                        "timeout": "20",
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                    },
                    {
                        "name": "migrate_from",
                        "timeout": "20",
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                    },
                    {
                        "name": "meta-data",
                        "timeout": "5",
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                    },
                    {
                        "name": "validate-all",
                        "timeout": "20",
                        "automatic": False,
                        "interval": None,
                        "on_target": False,
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                    },
                ],
                "default_actions": [
                    {
                        "name": "start",
                        "timeout": "20",
                        "interval": "0s",
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "stop",
                        "timeout": "20",
                        "interval": "0s",
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "monitor",
                        "timeout": "20",
                        "interval": "10",
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "reload",
                        "timeout": "20",
                        "interval": "0s",
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "migrate_to",
                        "timeout": "20",
                        "interval": "0s",
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                    {
                        "name": "migrate_from",
                        "timeout": "20",
                        "interval": "0s",
                        "role": None,
                        "start-delay": None,
                        "OCF_CHECK_LEVEL": None,
                        "automatic": False,
                        "on_target": False,
                    },
                ],
            },
        )
        self.env_assist.assert_reports(
            [
                fixture.info(
                    report_codes.AGENT_NAME_GUESSED,
                    entered_name="dummy",
                    guessed_name="ocf:heartbeat:Dummy",
                )
            ]
        )

    def test_agent_ambiguous(self):
        self.config.runner.pcmk.list_agents_standards("\n".join(["ocf"]))
        self.config.runner.pcmk.list_agents_ocf_providers(
            "\n".join(["heartbeat", "pacemaker"])
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:heartbeat",
            "\n".join(["agent1", "Dummy", "agent2"]),
            name="runner.pcmk.list_agents_ocf_providers_heartbeat",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:pacemaker",
            "\n".join(["agent1", "Dummy", "agent2"]),
            name="runner.pcmk.list_agents_ocf_providers_pacemaker",
        )
        self.env_assist.assert_raise_library_error(
            lambda: lib.describe_agent(self.env_assist.get_env(), "dummy")
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.AGENT_NAME_GUESS_FOUND_MORE_THAN_ONE,
                    agent="dummy",
                    possible_agents=[
                        "ocf:heartbeat:Dummy",
                        "ocf:pacemaker:Dummy",
                    ],
                )
            ],
        )

    def test_agent_not_found(self):
        self.config.runner.pcmk.list_agents_standards("\n".join(["ocf"]))
        self.config.runner.pcmk.list_agents_ocf_providers(
            "\n".join(["heartbeat"])
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:heartbeat", "\n".join(["Dummy"])
        )
        self.env_assist.assert_raise_library_error(
            lambda: lib.describe_agent(self.env_assist.get_env(), "agent")
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.AGENT_NAME_GUESS_FOUND_NONE,
                    agent="agent",
                )
            ],
        )

    def test_metadata_load_error(self):
        self.config.runner.pcmk.load_agent(
            agent_is_missing=True,
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
        )
        self.env_assist.assert_raise_library_error(
            lambda: lib.describe_agent(
                self.env_assist.get_env(), "ocf:heartbeat:Dummy"
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.UNABLE_TO_GET_AGENT_METADATA,
                    agent="ocf:heartbeat:Dummy",
                    reason=(
                        "Agent ocf:heartbeat:Dummy not found or does not support "
                        "meta-data: Invalid argument (22)\nMetadata query for "
                        "ocf:heartbeat:Dummy failed: Input/output error"
                    ),
                )
            ],
        )

    def test_invalid_name(self):
        self.env_assist.assert_raise_library_error(
            lambda: lib.describe_agent(
                self.env_assist.get_env(), "ocf:heartbeat:Something:else"
            )
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.INVALID_RESOURCE_AGENT_NAME,
                    name="ocf:heartbeat:Something:else",
                )
            ],
        )


class GetAgentMetadata(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)
        standard = "ocf"
        provider = "heartbeat"
        agent_type = "Dummy"
        self.name = ResourceAgentName(standard, provider, agent_type)
        self.agent_metadata = ResourceAgentMetadataDto(
            name=ResourceAgentNameDto(
                standard=standard,
                provider=provider,
                type=agent_type,
            ),
            shortdesc="Example stateless resource agent: ®",
            longdesc=(
                "This is a Dummy Resource Agent for testing utf-8"
                " in metadata: ®"
            ),
            parameters=[
                ResourceAgentParameterDto(
                    name="state-®",
                    shortdesc="State file: ®",
                    longdesc="Location to store the resource state in: ®",
                    type="string",
                    default="/var/run/resource-agents/Dummy-®.state",
                    enum_values=None,
                    required=False,
                    advanced=False,
                    deprecated=False,
                    deprecated_by=[],
                    deprecated_desc=None,
                    unique_group="_pcs_unique_group_state-®",
                    reloadable=True,
                ),
                ResourceAgentParameterDto(
                    name="trace_ra",
                    shortdesc=(
                        "Set to 1 to turn on resource agent "
                        "tracing (expect large output)"
                    ),
                    longdesc=(
                        "Set to 1 to turn on resource agent tracing"
                        " (expect large output) The trace output will be "
                        "saved to trace_file, if set, or by default to "
                        "$HA_VARRUN/ra_trace/<type>/<id>.<action>."
                        "<timestamp> e.g. $HA_VARRUN/ra_trace/oracle/db."
                        "start.2012-11-27.08:37:08"
                    ),
                    type="integer",
                    default="0",
                    enum_values=None,
                    required=False,
                    advanced=True,
                    deprecated=False,
                    deprecated_by=[],
                    deprecated_desc=None,
                    unique_group=None,
                    reloadable=False,
                ),
                ResourceAgentParameterDto(
                    name="trace_file",
                    shortdesc="Path to a file to store resource agent tracing log",
                    longdesc="Path to a file to store resource agent tracing log",
                    type="string",
                    default="",
                    enum_values=None,
                    required=False,
                    advanced=True,
                    deprecated=False,
                    deprecated_by=[],
                    deprecated_desc=None,
                    unique_group=None,
                    reloadable=False,
                ),
            ],
            actions=[
                ResourceAgentActionDto(
                    name="start",
                    timeout="20",
                    interval=None,
                    role=None,
                    start_delay=None,
                    depth=None,
                    automatic=False,
                    on_target=False,
                ),
                ResourceAgentActionDto(
                    name="stop",
                    timeout="20",
                    interval=None,
                    role=None,
                    start_delay=None,
                    depth=None,
                    automatic=False,
                    on_target=False,
                ),
                ResourceAgentActionDto(
                    name="monitor",
                    timeout="20",
                    interval="10",
                    role=None,
                    start_delay=None,
                    depth="0",
                    automatic=False,
                    on_target=False,
                ),
                ResourceAgentActionDto(
                    name="meta-data",
                    timeout="5",
                    interval=None,
                    role=None,
                    start_delay=None,
                    depth=None,
                    automatic=False,
                    on_target=False,
                ),
                ResourceAgentActionDto(
                    name="validate-all",
                    timeout="20",
                    interval=None,
                    role=None,
                    start_delay=None,
                    depth=None,
                    automatic=False,
                    on_target=False,
                ),
                ResourceAgentActionDto(
                    name="custom-®",
                    timeout="20",
                    interval=None,
                    role=None,
                    start_delay=None,
                    depth=None,
                    automatic=False,
                    on_target=False,
                ),
            ],
        )

    def test_full_name(self):
        self.config.runner.pcmk.load_agent(
            agent_name=self.name.full_name,
            agent_filename="resource_agent_ocf_heartbeat_dummy_utf8.xml",
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
        )
        self.assertEqual(
            lib.get_agent_metadata(self.env_assist.get_env(), self.name),
            self.agent_metadata,
        )

    def test_agent_not_found(self):
        err_msg = "error message"
        self.config.runner.pcmk.load_agent(
            agent_name=self.name.full_name,
            agent_is_missing=True,
            stderr=err_msg,
        )
        self.env_assist.assert_raise_library_error(
            lambda: lib.get_agent_metadata(self.env_assist.get_env(), self.name)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.UNABLE_TO_GET_AGENT_METADATA,
                    agent=self.name.full_name,
                    reason=err_msg,
                )
            ],
        )

    def test_metadata_load_error(self):
        self.config.runner.pcmk.load_agent(
            agent_is_missing=True,
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
        )
        self.env_assist.assert_raise_library_error(
            lambda: lib.get_agent_metadata(self.env_assist.get_env(), self.name)
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.UNABLE_TO_GET_AGENT_METADATA,
                    agent=self.name.full_name,
                    reason=(
                        f"Agent {self.name.full_name} not found or does not "
                        "support meta-data: Invalid argument (22)\nMetadata "
                        f"query for {self.name.full_name} failed: Input/output "
                        "error"
                    ),
                )
            ],
        )


class GetAgentsList(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_success(self):
        self.config.runner.pcmk.list_agents_standards(
            "\n".join(["service", "ocf"])
        )
        self.config.runner.pcmk.list_agents_ocf_providers("\n".join(["test"]))
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "ocf:test",
            "\n".join(["Stateful", "Delay"]),
            name="runner.pcmk.list_agents_ocf_providers.ocf_test",
        )
        self.config.runner.pcmk.list_agents_for_standard_and_provider(
            "service",
            "\n".join(["corosync", "pacemaker_remote"]),
            name="runner.pcmk.list_agents_ocf_providers.service",
        )
        self.assertEqual(
            lib.get_agents_list(self.env_assist.get_env()),
            ListResourceAgentNameDto(
                names=[
                    ResourceAgentNameDto(
                        standard="ocf",
                        provider="test",
                        type="Delay",
                    ),
                    ResourceAgentNameDto(
                        standard="ocf",
                        provider="test",
                        type="Stateful",
                    ),
                    ResourceAgentNameDto(
                        standard="service",
                        provider=None,
                        type="corosync",
                    ),
                    ResourceAgentNameDto(
                        standard="service",
                        provider=None,
                        type="pacemaker_remote",
                    ),
                ]
            ),
        )


class GetAgentDefaultOperations(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(test_case=self)

    def test_resource(self):
        agent_name = ResourceAgentName("ocf", "pacemaker", "Stateful")
        self.config.runner.pcmk.load_agent(
            agent_name=agent_name.full_name,
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
        )
        self.assertEqual(
            lib.get_agent_default_operations(
                self.env_assist.get_env(), agent_name.to_dto()
            ),
            ListCibResourceOperationDto(
                operations=[
                    _operation_fixture("start", "0s", timeout="20s"),
                    _operation_fixture("stop", "0s", timeout="20s"),
                    _operation_fixture(
                        "monitor",
                        "10s",
                        timeout="20s",
                        role=const.PCMK_ROLE_PROMOTED_PRIMARY,
                    ),
                    _operation_fixture(
                        "monitor",
                        "11s",
                        timeout="20s",
                        role=const.PCMK_ROLE_UNPROMOTED_PRIMARY,
                    ),
                    _operation_fixture("promote", "0s", timeout="10s"),
                    _operation_fixture("demote", "0s", timeout="10s"),
                    _operation_fixture("notify", "0s", timeout="5s"),
                    _operation_fixture("reload-agent", "0s", timeout="10s"),
                ]
            ),
        )

    def test_resource_only_necessary(self):
        agent_name = ResourceAgentName("ocf", "pacemaker", "Stateful")
        self.config.runner.pcmk.load_agent(
            agent_name=agent_name.full_name,
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
        )
        self.assertEqual(
            lib.get_agent_default_operations(
                self.env_assist.get_env(),
                agent_name.to_dto(),
                necessary_only=True,
            ),
            ListCibResourceOperationDto(
                operations=[
                    _operation_fixture(
                        "monitor",
                        "10s",
                        timeout="20s",
                        role=const.PCMK_ROLE_PROMOTED_PRIMARY,
                    ),
                    _operation_fixture(
                        "monitor",
                        "11s",
                        timeout="20s",
                        role=const.PCMK_ROLE_UNPROMOTED_PRIMARY,
                    ),
                ]
            ),
        )

    def _test_stonith(self, necessary_only):
        agent_name = ResourceAgentName("stonith", None, "fence_unfencing")
        self.config.runner.pcmk.load_agent(
            agent_name=agent_name.full_name,
            env={"PATH": "/usr/sbin:/bin:/usr/bin"},
        )
        self.config.runner.pcmk.load_fenced_metadata()
        self.assertEqual(
            lib.get_agent_default_operations(
                self.env_assist.get_env(),
                agent_name.to_dto(),
                necessary_only=necessary_only,
            ),
            ListCibResourceOperationDto(
                operations=[
                    _operation_fixture("monitor", interval="60s"),
                ]
            ),
        )

    def test_stonith(self):
        self._test_stonith(False)

    def test_stonith_only_necessary(self):
        self._test_stonith(True)



from typing import (
    List,
    Optional,
    cast,
)

from lxml.etree import (
    SubElement,
    _Element,
)

from pcs.common.pacemaker.resource.group import CibResourceGroupDto
from pcs.lib.cib import (
    nvpair_multi,
    rule,
)
from pcs.lib.cib.const import TAG_RESOURCE_GROUP as TAG


def is_group(resource_el: _Element) -> bool:
    return resource_el.tag == TAG


def append_new(resources_section: _Element, group_id: str) -> _Element:
    return SubElement(resources_section, TAG, id=group_id)


def get_inner_resources(
    group_el: _Element,
) -> List[_Element]:
    return cast(List[_Element], group_el.xpath("./primitive"))


def group_element_to_dto(
    group_element: _Element,
    rule_eval: Optional[rule.RuleInEffectEval] = None,
) -> CibResourceGroupDto:
    if rule_eval is None:
        rule_eval = rule.RuleInEffectEvalDummy()
    return CibResourceGroupDto(
        id=str(group_element.attrib["id"]),
        description=group_element.get("description"),
        member_ids=[
            str(primitive_el.attrib["id"])
            for primitive_el in get_inner_resources(group_element)
        ],
        meta_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                group_element, nvpair_multi.NVSET_META
            )
        ],
        instance_attributes=[
            nvpair_multi.nvset_element_to_dto(nvset, rule_eval)
            for nvset in nvpair_multi.find_nvsets(
                group_element, nvpair_multi.NVSET_INSTANCE
            )
        ],
    )



from .. import const


def get_value_for_cib(
    role: const.PcmkRoleType, is_latest_supported: bool
) -> const.PcmkRoleType:
    if is_latest_supported:
        return get_value_primary(role)
    if role in const.PCMK_ROLES_PROMOTED:
        return const.PCMK_ROLE_PROMOTED_LEGACY
    if role in const.PCMK_ROLES_UNPROMOTED:
        return const.PCMK_ROLE_UNPROMOTED_LEGACY
    return role


def get_value_primary(role: const.PcmkRoleType) -> const.PcmkRoleType:
    if role in const.PCMK_ROLES_PROMOTED:
        return const.PCMK_ROLE_PROMOTED_PRIMARY
    if role in const.PCMK_ROLES_UNPROMOTED:
        return const.PCMK_ROLE_UNPROMOTED_PRIMARY
    return role



import os
from textwrap import dedent
from unittest import (
    TestCase,
    mock,
)

from pcs.cli.booth import command as booth_cmd

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import get_test_resource as rc
from pcs_test.tools.misc import (
    get_tmp_dir,
    get_tmp_file,
    outdent,
    skip_unless_booth_resource_agent_installed,
    write_file_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

EMPTY_CIB = rc("cib-empty.xml")


class BoothMixinNoFiles(AssertPcsMixin):
    def setUp(self):
        # pylint cannot possibly know this is being mixed into TestCase classes
        # pylint: disable=invalid-name
        self.pcs_runner = PcsRunner(None)


class BoothMixin(AssertPcsMixin):
    def setUp(self):
        # pylint cannot possibly know this is being mixed into TestCase classes
        # pylint: disable=invalid-name
        self.booth_dir = get_tmp_dir("tier1_booth")
        self.booth_cfg_path = os.path.join(self.booth_dir.name, "booth.cfg")
        self.booth_key_path = os.path.join(self.booth_dir.name, "booth.key")
        self.temp_cib = get_tmp_file("tier1_booth")
        write_file_to_tmpfile(EMPTY_CIB, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        # pylint cannot possibly know this is being mixed into TestCase classes
        # pylint: disable=invalid-name
        self.temp_cib.close()
        self.booth_dir.cleanup()

    def fake_file(self, command):
        return command + [
            f"--booth-conf={self.booth_cfg_path}",
            f"--booth-key={self.booth_key_path}",
        ]

    def ensure_booth_config_exists(self):
        if not os.path.exists(self.booth_cfg_path):
            with open(self.booth_cfg_path, "w") as config_file:
                config_file.write("")

    def ensure_booth_config_not_exists(self):
        if os.path.exists(self.booth_cfg_path):
            os.remove(self.booth_cfg_path)
        if os.path.exists(self.booth_key_path):
            os.remove(self.booth_key_path)

    def assert_pcs_success(self, command, *args, **kwargs):
        # pylint: disable=signature-differs
        return super().assert_pcs_success(
            self.fake_file(command), *args, **kwargs
        )

    def assert_pcs_fail(self, command, *args, **kwargs):
        # pylint: disable=signature-differs
        return super().assert_pcs_fail(self.fake_file(command), *args, **kwargs)

    def assert_pcs_fail_original(self, *args, **kwargs):
        return super().assert_pcs_fail(*args, **kwargs)


class SetupTest(BoothMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.pcs_runner.cib_file = None

    def test_success_setup_booth_config(self):
        self.ensure_booth_config_not_exists()
        self.assert_pcs_success(
            "booth setup sites 1.1.1.1 2.2.2.2 arbitrators 3.3.3.3".split()
        )
        with open(self.booth_cfg_path, "r") as config_file:
            self.assertEqual(
                dedent(
                    """\
                    authfile = {0}
                    site = 1.1.1.1
                    site = 2.2.2.2
                    arbitrator = 3.3.3.3
                    """.format(
                        self.booth_key_path
                    )
                ),
                config_file.read(),
            )
        with open(self.booth_key_path, "rb") as key_file:
            self.assertEqual(64, len(key_file.read()))

    def test_overwrite_existing_mocked_config(self):
        self.ensure_booth_config_exists()
        self.assert_pcs_success(
            "booth setup sites 1.1.1.1 2.2.2.2 arbitrators 3.3.3.3".split(),
        )
        self.ensure_booth_config_not_exists()

    def test_fail_on_multiple_reasons(self):
        self.assert_pcs_fail(
            (
                "booth setup sites 1.1.1.1 arbitrators 1.1.1.1 2.2.2.2 3.3.3.3"
            ).split(),
            (
                "Error: lack of sites for booth configuration (need 2 at least)"
                ": sites '1.1.1.1'\n"
                "Error: odd number of peers is required (entered 4 peers)\n"
                "Error: duplicate address for booth configuration: '1.1.1.1'\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )

    def test_refuse_partially_mocked_environment(self):
        self.assert_pcs_fail_original(
            # no --booth-key!
            (
                "booth setup sites 1.1.1.1 2.2.2.2 arbitrators 3.3.3.3 "
                "--booth-conf=/some/file"
            ).split(),
            (
                "Error: When --booth-conf is specified, --booth-key must be "
                "specified as well\n"
            ),
        )
        self.assert_pcs_fail_original(
            # no --booth-conf!
            (
                "booth setup sites 1.1.1.1 2.2.2.2 arbitrators 3.3.3.3 "
                "--booth-key=/some/file"
            ).split(),
            (
                "Error: When --booth-key is specified, --booth-conf must be "
                "specified as well\n"
            ),
        )

    def test_show_usage_when_no_site_specified(self):
        self.assert_pcs_fail(
            "booth setup arbitrators 3.3.3.3".split(),
            stdout_start="\nUsage: pcs booth <command>\n    setup",
        )
        self.assert_pcs_fail(
            "booth setup".split(),
            stdout_start="\nUsage: pcs booth <command>\n    setup",
        )


class DestroyTest(BoothMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.pcs_runner.cib_file = None

    def test_failed_when_using_mocked_booth_env(self):
        self.assert_pcs_fail(
            "booth destroy".split(),
            (
                "Error: Specified options '--booth-conf', '--booth-key' are "
                "not supported in this command\n"
            ),
        )


class BoothTest(BoothMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.pcs_runner.cib_file = None
        self.ensure_booth_config_not_exists()
        self.assert_pcs_success(
            "booth setup sites 1.1.1.1 2.2.2.2 arbitrators 3.3.3.3".split()
        )


class AddTicketTest(BoothTest):
    def test_success_add_ticket(self):
        self.assert_pcs_success("booth ticket add TicketA expire=10".split())
        with open(self.booth_cfg_path, "r") as config_file:
            self.assertEqual(
                dedent(
                    """\
                    authfile = {0}
                    site = 1.1.1.1
                    site = 2.2.2.2
                    arbitrator = 3.3.3.3
                    ticket = "TicketA"
                      expire = 10
                    """.format(
                        self.booth_key_path
                    )
                ),
                config_file.read(),
            )

    def test_fail_on_bad_ticket_name(self):
        self.assert_pcs_fail(
            "booth ticket add @TicketA".split(),
            (
                "Error: booth ticket name '@TicketA' is not valid, use up to 63 "
                "alphanumeric characters or dash\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )

    def test_fail_on_duplicit_ticket_name(self):
        self.assert_pcs_success("booth ticket add TicketA".split())
        self.assert_pcs_fail(
            "booth ticket add TicketA".split(),
            (
                "Error: booth ticket name 'TicketA' already exists in "
                "configuration\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )

    def test_fail_on_invalid_options(self):
        self.assert_pcs_fail(
            "booth ticket add TicketA site=a timeout=".split(),
            (
                "Error: invalid booth ticket option 'site', allowed options"
                " are: 'acquire-after', 'attr-prereq', "
                "'before-acquire-handler', 'expire', 'mode', 'renewal-freq', "
                "'retries', 'timeout', 'weights'\n"
                "Error: timeout cannot be empty\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )

    def test_forceable_fail_on_unknown_options(self):
        msg = (
            "invalid booth ticket option 'unknown', allowed options"
            " are: 'acquire-after', 'attr-prereq', 'before-acquire-handler',"
            " 'expire', 'mode', 'renewal-freq', 'retries', 'timeout', 'weights'"
        )
        self.assert_pcs_fail(
            "booth ticket add TicketA unknown=a".split(),
            (
                "Error: {0}, use --force to override\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ).format(msg),
        )
        self.assert_pcs_success(
            "booth ticket add TicketA unknown=a --force".split(),
            "Warning: {0}\n".format(msg),
        )

    def test_not_enough_args(self):
        self.assert_pcs_fail(
            "booth ticket add".split(),
            stdout_start="\nUsage: pcs booth <command>\n    ticket add",
        )


class DeleteRemoveTicketMixin:
    command = None

    def test_not_enough_args(self):
        self.assert_pcs_fail(
            ["booth", "ticket", self.command],
            stdout_start=outdent(
                f"""
                Usage: pcs booth <command>
                    ticket {self.command} <"""
            ),
        )

    def test_too_many_args(self):
        self.assert_pcs_fail(
            ["booth", "ticket", self.command, "aaa", "bbb"],
            stdout_start=outdent(
                f"""
                Usage: pcs booth <command>
                    ticket {self.command} <"""
            ),
        )

    def test_success_remove_ticket(self):
        self.assert_pcs_success("booth ticket add TicketA".split())
        with open(self.booth_cfg_path, "r") as config_file:
            self.assertEqual(
                dedent(
                    """\
                    authfile = {0}
                    site = 1.1.1.1
                    site = 2.2.2.2
                    arbitrator = 3.3.3.3
                    ticket = "TicketA"
                    """.format(
                        self.booth_key_path
                    )
                ),
                config_file.read(),
            )
        self.assert_pcs_success(["booth", "ticket", self.command, "TicketA"])
        with open(self.booth_cfg_path, "r") as config_file:
            self.assertEqual(
                dedent(
                    """\
                    authfile = {0}
                    site = 1.1.1.1
                    site = 2.2.2.2
                    arbitrator = 3.3.3.3
                    """.format(
                        self.booth_key_path
                    )
                ),
                config_file.read(),
            )

    def test_fail_when_ticket_does_not_exist(self):
        self.assert_pcs_fail(
            ["booth", "ticket", self.command, "TicketA"],
            (
                "Error: booth ticket name 'TicketA' does not exist\n"
                "Error: Errors have occurred, therefore pcs is unable to "
                "continue\n"
            ),
        )


class DeleteTicketTest(DeleteRemoveTicketMixin, BoothTest):
    command = "delete"


class RemoveTicketTest(DeleteRemoveTicketMixin, BoothTest):
    command = "remove"


@skip_unless_booth_resource_agent_installed()
class CreateTest(BoothMixinNoFiles, TestCase):
    def test_not_enough_args(self):
        self.assert_pcs_fail(
            "booth create".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    create ip <"""
            ),
        )
        self.assert_pcs_fail(
            "booth create ip".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    create ip <"""
            ),
        )

    def test_too_many_args(self):
        self.assert_pcs_fail(
            "booth create ip aaa bbb".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    create ip <"""
            ),
        )


class DeleteRemoveTestMixin(AssertPcsMixin):
    command = None

    def setUp(self):
        # pylint cannot know this will be mixed into a TetsCase class
        # pylint: disable=invalid-name
        self.temp_cib = get_tmp_file("tier1_booth_delete_remove")
        write_file_to_tmpfile(EMPTY_CIB, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib.name)

    def tearDown(self):
        # pylint cannot possibly know this is being mixed into TestCase classes
        # pylint: disable=invalid-name
        self.temp_cib.close()

    def test_usage(self):
        self.assert_pcs_fail(
            ["booth", self.command, "a", "b"],
            stdout_start=outdent(
                f"""
                Usage: pcs booth <command>
                    {self.command}
                """
            ),
        )

    def test_failed_when_no_booth_configuration_created(self):
        self.assert_pcs_success(
            "resource status".split(), "NO resources configured\n"
        )
        self.assert_pcs_fail(
            ["booth", self.command],
            [
                # pylint: disable=line-too-long
                "Error: booth instance 'booth' not found in cib",
                "Error: Errors have occurred, therefore pcs is unable to continue",
            ],
        )


@skip_unless_booth_resource_agent_installed()
class DeleteTest(DeleteRemoveTestMixin, TestCase):
    command = "delete"


@skip_unless_booth_resource_agent_installed()
class RemoveTest(DeleteRemoveTestMixin, TestCase):
    command = "remove"


class TicketGrantTest(BoothMixinNoFiles, TestCase):
    def test_not_enough_args(self):
        self.assert_pcs_fail(
            "booth ticket grant".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    ticket grant <"""
            ),
        )

    def test_too_many_args(self):
        self.assert_pcs_fail(
            "booth ticket grant aaa bbb ccc".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    ticket grant <"""
            ),
        )


class TicketRevokeTest(BoothMixinNoFiles, TestCase):
    def test_not_enough_args(self):
        self.assert_pcs_fail(
            "booth ticket revoke".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    ticket revoke <"""
            ),
        )

    def test_too_many_args(self):
        self.assert_pcs_fail(
            "booth ticket revoke aaa bbb ccc".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    ticket revoke <"""
            ),
        )


class Restart(BoothMixinNoFiles, TestCase):
    def test_too_many_args(self):
        self.assert_pcs_fail(
            "booth restart aaa".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    restart"""
            ),
        )


class Sync(BoothMixinNoFiles, TestCase):
    def test_too_many_args(self):
        self.assert_pcs_fail(
            "booth sync aaa".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    sync"""
            ),
        )


class BoothServiceTestMixin(BoothMixinNoFiles):
    def test_too_many_args(self):
        self.assert_pcs_fail(
            ["booth", self.cmd_label, "aaa"],
            stdout_start=outdent(
                f"""
                Usage: pcs booth <command>
                    {self.cmd_label}"""
            ),
        )


class Enable(BoothServiceTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.cmd_label = "enable"
        self.cli_cmd = booth_cmd.enable


class Disable(BoothServiceTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.cmd_label = "disable"
        self.cli_cmd = booth_cmd.disable


class Start(BoothServiceTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.cmd_label = "start"
        self.cli_cmd = booth_cmd.start


class Stop(BoothServiceTestMixin, TestCase):
    def setUp(self):
        super().setUp()
        self.cmd_label = "stop"
        self.cli_cmd = booth_cmd.stop


class Pull(BoothMixinNoFiles, TestCase):
    def test_not_enough_args(self):
        self.assert_pcs_fail(
            "booth pull".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    pull"""
            ),
        )

    def test_too_many_args(self):
        self.assert_pcs_fail(
            "booth pull aaa bbb".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    pull"""
            ),
        )


# disable printig the booth status so it won't break tests output
@mock.patch("pcs.cli.booth.command.print", new=lambda x: x)
class Status(BoothMixinNoFiles, TestCase):
    def test_too_many_args(self):
        self.assert_pcs_fail(
            "booth status aaa".split(),
            stdout_start=outdent(
                """
                Usage: pcs booth <command>
                    status"""
            ),
        )



import json
from textwrap import dedent
from unittest import TestCase

from pcs_test.tools.assertions import AssertPcsMixin
from pcs_test.tools.misc import (
    get_tmp_file,
    write_data_to_tmpfile,
)
from pcs_test.tools.pcs_runner import PcsRunner

from .common import fixture_corosync_conf_minimal


def fixture_text_output(no_cluster_uuid=False):
    text_output = "Cluster Name: cluster_name\n"
    if not no_cluster_uuid:
        text_output += "Cluster UUID: cluster_uuid\n"
    text_output += dedent(
        """\
        Transport: knet
        Nodes:
          node1:
            Link 0 address: node1_addr
            nodeid: 1
          node2:
            Link 0 address: node2_addr
            nodeid: 2
        Transport Options:
          ip_version: ipv6
        Crypto Options:
          cipher: aes256
          hash: sha256
      """
    )
    return text_output


def fixture_json_output(no_cluster_uuid=False):
    return (
        json.dumps(
            {
                "cluster_name": "cluster_name",
                "cluster_uuid": None if no_cluster_uuid else "cluster_uuid",
                "transport": "KNET",
                "totem_options": {},
                "transport_options": {"ip_version": "ipv6"},
                "compression_options": {},
                "crypto_options": {"cipher": "aes256", "hash": "sha256"},
                "nodes": [
                    {
                        "name": "node1",
                        "nodeid": "1",
                        "addrs": [
                            {
                                "addr": "node1_addr",
                                "link": "0",
                                "type": "FQDN",
                            }
                        ],
                    },
                    {
                        "name": "node2",
                        "nodeid": "2",
                        "addrs": [
                            {
                                "addr": "node2_addr",
                                "link": "0",
                                "type": "FQDN",
                            }
                        ],
                    },
                ],
                "links_options": {},
                "quorum_options": {},
                "quorum_device": None,
            }
        )
        + "\n"
    )


def fixture_cmd_output(no_cluster_uuid=False):
    cmd_output = dedent(
        """\
        pcs cluster setup cluster_name \\
          node1 addr=node1_addr \\
          node2 addr=node2_addr \\
          transport \\
          knet \\
              ip_version=ipv6 \\
            crypto \\
              cipher=aes256 \\
              hash=sha256"""
    )
    if no_cluster_uuid:
        cmd_output += " \\\n  --no-cluster-uuid"

    return cmd_output + "\n"


class ClusterConfigMixin(AssertPcsMixin):
    # pylint: disable=invalid-name
    command = None

    def setUp(self):
        self.corosync_conf_file = get_tmp_file(
            "tier1_cluster_config_show_corosync.conf"
        )
        self.pcs_runner = PcsRunner(
            cib_file=None,
            corosync_conf_opt=self.corosync_conf_file.name,
        )
        write_data_to_tmpfile(
            fixture_corosync_conf_minimal(), self.corosync_conf_file
        )

    def tearDown(self):
        self.corosync_conf_file.close()

    def test_default_output(self):
        self.assert_pcs_success(
            self.command.split(),
            stdout_full=fixture_text_output(),
        )

    def test_text_output(self):
        self.assert_pcs_success(
            (self.command + " --output-format=text").split(),
            stdout_full=fixture_text_output(),
        )

    def test_json_output(self):
        self.assert_pcs_success(
            (self.command + " --output-format=json").split(),
            stdout_full=fixture_json_output(),
        )

    def test_cmd_output(self):
        self.assert_pcs_success(
            (self.command + " --output-format=cmd").split(),
            stdout_full=fixture_cmd_output(),
        )

    def test_output_format_unsupported_value(self):
        self.assert_pcs_fail(
            (self.command + " --output-format=xml").split(),
            stdout_full=(
                "Error: Unknown value 'xml' for '--output-format' option. "
                "Supported values are: 'cmd', 'json', 'text'\n"
            ),
        )

    def test_unsupported_option(self):
        self.assert_pcs_fail(
            (self.command + " --corosync").split(),
            stdout_full=dedent(
                """\
                Error: Specified option '--corosync' is not supported in this command
                """
            ),
        )


class ClusterConfig(ClusterConfigMixin, TestCase):
    command = "cluster config"


class ClusterConfigShow(ClusterConfigMixin, TestCase):
    command = "cluster config show"


class ClusterConfigNoUuid(AssertPcsMixin, TestCase):
    # pylint: disable=invalid-name
    def setUp(self):
        self.corosync_conf_file_no_uuid = get_tmp_file(
            "tier1_cluster_config_show_corosync_no_uuid.conf"
        )
        self.pcs_runner = PcsRunner(
            cib_file=None,
            corosync_conf_opt=self.corosync_conf_file_no_uuid.name,
        )
        write_data_to_tmpfile(
            fixture_corosync_conf_minimal(no_cluster_uuid=True),
            self.corosync_conf_file_no_uuid,
        )

    def tearDown(self):
        self.corosync_conf_file_no_uuid.close()

    def test_default_output(self):
        self.assert_pcs_success(
            ["cluster", "config", "show"],
            stdout_full=fixture_text_output(no_cluster_uuid=True),
        )

    def test_text_output(self):
        self.assert_pcs_success(
            ["cluster", "config", "show", "--output-format=text"],
            stdout_full=fixture_text_output(no_cluster_uuid=True),
        )

    def test_json_output_no_uuid(self):
        self.assert_pcs_success(
            ["cluster", "config", "show", "--output-format=json"],
            stdout_full=fixture_json_output(no_cluster_uuid=True),
        )

    def test_cmd_output_no_uuid(self):
        self.assert_pcs_success(
            ["cluster", "config", "show", "--output-format=cmd"],
            stdout_full=fixture_cmd_output(no_cluster_uuid=True),
        )



from typing import (
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    cast,
)

from pcs import settings
from pcs.common import reports
from pcs.common.reports.item import ReportItem
from pcs.lib.corosync import (
    config_parser,
    constants,
    node,
)
from pcs.lib.errors import LibraryError
from pcs.lib.interface.config import FacadeInterface

_KNET_COMPRESSION_OPTIONS_PREFIX = "knet_compression_"
_KNET_CRYPTO_OPTIONS_PREFIX = "crypto_"


class ConfigFacade(FacadeInterface):
    # pylint: disable=too-many-public-methods
    """
    Provides high level access to a corosync config file
    """

    @classmethod
    def create(cls, cluster_name, node_list, transport):
        """
        Create a minimal config

        string cluster_name -- a name of a cluster
        list node_list -- list of dict: name, addrs
        string transport -- corosync transport
        """
        root = config_parser.Section("")
        totem_section = config_parser.Section("totem")
        nodelist_section = config_parser.Section("nodelist")
        quorum_section = config_parser.Section("quorum")
        logging_section = config_parser.Section("logging")
        root.add_section(totem_section)
        root.add_section(nodelist_section)
        root.add_section(quorum_section)
        root.add_section(logging_section)

        totem_section.add_attribute("version", "2")
        totem_section.add_attribute("cluster_name", cluster_name)
        totem_section.add_attribute("transport", transport)
        quorum_section.add_attribute("provider", "corosync_votequorum")
        logging_section.add_attribute("to_logfile", "yes")
        logging_section.add_attribute("logfile", settings.corosync_log_file)
        logging_section.add_attribute("to_syslog", "yes")
        logging_section.add_attribute("timestamp", "on")

        for node_id, node_options in enumerate(node_list, 1):
            nodelist_section.add_section(
                cls._create_node_section(
                    node_id, node_options, range(constants.LINKS_MAX)
                )
            )

        self = cls(root)
        # pylint: disable=protected-access
        self.__update_two_node()

        return self

    def __init__(self, parsed_config):
        """
        Create a facade around a parsed corosync config file
        parsed_config parsed corosync config
        """
        super().__init__(parsed_config)
        # set to True if changes cannot be applied on running cluster
        self._need_stopped_cluster = False
        # set to True if qdevice reload is required to apply changes
        self._need_qdevice_reload = False

    @property
    def need_stopped_cluster(self):
        return self._need_stopped_cluster

    @property
    def need_qdevice_reload(self):
        return self._need_qdevice_reload

    def get_cluster_name(self) -> str:
        return cast(str, self._get_option_value("totem", "cluster_name", ""))

    def get_cluster_uuid(self) -> Optional[str]:
        return self._get_option_value("totem", "cluster_uuid")

    def set_cluster_uuid(self, cluster_uuid: str) -> None:
        """
        Updates or adds a cluster_uuid in the totem section, assumes that UUID
        can be rewritten

            cluster_uuid - new cluster UUID
        """
        totem_section_list = self.__ensure_section(self.config, "totem")
        self.__set_section_options(
            totem_section_list, {"cluster_uuid": cluster_uuid}
        )
        self.__remove_empty_sections(self.config)

    # To get a list of nodenames use pcs.lib.node.get_existing_nodes_names

    def get_nodes(self) -> List[node.CorosyncNode]:
        """
        Get all defined nodes
        """
        result = []
        for nodelist in self.config.get_sections("nodelist"):
            for node_section in nodelist.get_sections("node"):
                # first, load all the nodes key-value pairs so that the last
                # value for each key wins
                node_data = self._get_node_data(node_section)
                if not node_data:
                    continue
                # add the node data to the resulting list
                result.append(
                    node.CorosyncNode(
                        node_data.get("name"),
                        [
                            node.CorosyncNodeAddress(
                                node_data[f"ring{i}_addr"], str(i)
                            )
                            for i in range(constants.LINKS_MAX)
                            if node_data.get(f"ring{i}_addr")
                        ],
                        node_data.get("nodeid"),
                    )
                )
        return result

    def _get_used_nodeid_list(self):
        used_ids = []
        for nodelist in self.config.get_sections("nodelist"):
            for node_section in nodelist.get_sections("node"):
                used_ids.extend(
                    [
                        int(attr[1])
                        for attr in node_section.get_attributes("nodeid")
                    ]
                )
        return used_ids

    @staticmethod
    def _get_nodeid_generator(used_ids):
        used_ids = set(used_ids)
        current_id = 1
        while True:
            if current_id not in used_ids:
                yield current_id
                used_ids.add(current_id)
            current_id += 1

    @staticmethod
    def _get_node_data(node_section):
        return {
            attr_name: attr_value
            for attr_name, attr_value in node_section.get_attributes()
            if attr_name in constants.NODE_OPTIONS
        }

    def get_used_linknumber_list(self):
        for nodelist_section in self.config.get_sections("nodelist"):
            for node_section in nodelist_section.get_sections("node"):
                node_data = self._get_node_data(node_section)
                if not node_data:
                    continue
                return [
                    str(i)
                    for i in range(constants.LINKS_MAX)
                    if node_data.get(f"ring{i}_addr")
                ]

    @staticmethod
    def _create_node_section(node_id, node_options, link_ids):
        node_section = config_parser.Section("node")
        for link_id, link_addr in zip(link_ids, node_options["addrs"]):
            node_section.add_attribute("ring{}_addr".format(link_id), link_addr)
        node_section.add_attribute("name", node_options["name"])
        node_section.add_attribute("nodeid", node_id)
        return node_section

    def add_nodes(self, node_list):
        """
        Add nodes to a config with a nonempty nodelist

        list node_list -- list of dict: name, addrs
        """
        nodelist_section = self.__ensure_section(self.config, "nodelist")[-1]
        node_id_generator = self._get_nodeid_generator(
            self._get_used_nodeid_list()
        )
        for node_options in node_list:
            nodelist_section.add_section(
                self._create_node_section(
                    next(node_id_generator),
                    node_options,
                    self.get_used_linknumber_list(),
                )
            )
        self.__update_two_node()

    def remove_nodes(self, node_name_list):
        """
        Remove nodes from a config

        iterable node_name_list -- names of nodes to remove
        """
        for nodelist_section in self.config.get_sections("nodelist"):
            for node_section in nodelist_section.get_sections("node"):
                node_data = self._get_node_data(node_section)
                if node_data.get("name") in node_name_list:
                    nodelist_section.del_section(node_section)
        self.__remove_empty_sections(self.config)
        self.__update_two_node()

    def create_link_list(self, link_list):
        """
        Add a link list to a config without one

        iterable link_list -- list of dicts with link_list options
        """
        available_link_numbers = list(range(constants.LINKS_KNET_MAX))
        linknumber_missing = []
        links = []

        for link in link_list:
            if "linknumber" in link:
                try:
                    available_link_numbers.remove(int(link["linknumber"]))
                except ValueError as e:
                    raise AssertionError(
                        "Invalid link number: {}".format(e)
                    ) from e
                links.append(dict(link))
            else:
                linknumber_missing.append(link)

        for link in linknumber_missing:
            link = dict(link)
            try:
                link["linknumber"] = str(available_link_numbers.pop(0))
            except IndexError as e:
                raise AssertionError(
                    "Link number no longer available: {}".format(e)
                ) from e
            links.append(link)

        for link in sorted(links, key=lambda item: item["linknumber"]):
            self._set_link_options(link)

    def add_link(self, node_addr_map, options):
        """
        Add a new link to nodelist and create an interface section with options

        dict node_addr_map -- key: node name, value: node address for the link
        dict link_options -- link options
        """
        # Get a linknumber
        if "linknumber" in options:
            linknumber = options["linknumber"]
        else:
            linknumber = None
            used_links = self.get_used_linknumber_list()
            available_links = range(constants.LINKS_KNET_MAX)
            for candidate in available_links:
                if str(candidate) not in used_links:
                    linknumber = candidate
                    break
            if linknumber is None:
                raise AssertionError("No link number available")
            options["linknumber"] = linknumber

        # Add addresses
        for nodelist_section in self.config.get_sections("nodelist"):
            for node_section in nodelist_section.get_sections("node"):
                node_name = self._get_node_data(node_section).get("name")
                if node_name in node_addr_map:
                    node_section.add_attribute(
                        f"ring{linknumber}_addr", node_addr_map[node_name]
                    )

        # Add link options.
        if options:
            self._set_link_options(options)

    def _set_link_options(
        self, options, interface_section_list=None, linknumber=None
    ):
        """
        Add a new or change an existing interface section with link options

        dict options -- link options
        list interface_section_list -- list of existing sections to be changed
        string linknumber -- linknumber to set to a newly created section
        """
        # If the only option is "linknumber" then there is no point in adding
        # the options at all. It would mean there are no options for the
        # particular link.
        if not [name for name in options if name != "linknumber"]:
            return

        options_to_set = self.__translate_link_options(options)
        if not interface_section_list:
            new_section = config_parser.Section("interface")
            if linknumber:
                new_section.set_attribute("linknumber", linknumber)
            totem_section = self.__ensure_section(self.config, "totem")[-1]
            totem_section.add_section(new_section)
            interface_section_list = [new_section]
        self.__set_section_options(interface_section_list, options_to_set)
        self.__remove_empty_sections(self.config)

    def remove_links(self, link_list):
        """
        Remove links from nodelist and relevant interface sections from totem

        iterable link_list -- list of linknumbers (strings) to be removed
        """
        # Do not break when the interface / address is found to be sure to
        # remove all of them (config format allows to have more interface
        # sections / addresses for one link).
        for totem_section in self.config.get_sections("totem"):
            for interface_section in totem_section.get_sections("interface"):
                interface_number = interface_section.get_attribute_value(
                    "linknumber",
                    # if no linknumber is set, corosync treats it as 0
                    "0",
                )
                if interface_number in link_list:
                    totem_section.del_section(interface_section)
        for link_number in link_list:
            for nodelist_section in self.config.get_sections("nodelist"):
                for node_section in nodelist_section.get_sections("node"):
                    node_section.del_attributes_by_name(
                        f"ring{link_number}_addr"
                    )
        self.__remove_empty_sections(self.config)

    def update_link(self, linknumber, node_addr_map, options):
        """
        Change an existing link - node addresses and/or link options

        string linknumber -- link to be changed
        dict node_addr_map -- key: node name, value: node address for the link
        dict link_options -- link options
        """
        self._need_stopped_cluster = True
        # make sure we do not change the linknumber
        if "linknumber" in options:
            del options["linknumber"]
        # change addresses
        if node_addr_map:
            for nodelist_section in self.config.get_sections("nodelist"):
                for node_section in nodelist_section.get_sections("node"):
                    node_name = self._get_node_data(node_section).get("name")
                    if node_name in node_addr_map:
                        node_section.set_attribute(
                            f"ring{linknumber}_addr", node_addr_map[node_name]
                        )
        # change options
        if options:
            target_interface_section_list = []
            for totem_section in self.config.get_sections("totem"):
                for interface_section in totem_section.get_sections(
                    "interface"
                ):
                    if (
                        linknumber
                        ==
                        # if no linknumber is set, corosync treats it as 0
                        interface_section.get_attribute_value("linknumber", "0")
                    ):
                        target_interface_section_list.append(interface_section)
            self._set_link_options(
                options,
                interface_section_list=target_interface_section_list,
                linknumber=linknumber,
            )
        self.__remove_empty_sections(self.config)

    def get_links_options(self):
        """
        Get all links' options in a dict: key=linknumber value=dict of options
        """
        transport = self.get_transport()
        allowed_options = (
            constants.LINK_OPTIONS_UDP
            if transport in constants.TRANSPORTS_UDP
            else constants.LINK_OPTIONS_KNET_COROSYNC
        )
        raw_options = {}
        for totem_section in self.config.get_sections("totem"):
            for interface_section in totem_section.get_sections("interface"):
                # if no linknumber is set, corosync treats it as 0
                linknumber = interface_section.get_attribute_value(
                    "linknumber", "0"
                )
                if linknumber not in raw_options:
                    raw_options[linknumber] = {}
                for name, value in interface_section.get_attributes():
                    if name in allowed_options:
                        raw_options[linknumber][name] = value
                # make sure the linknumber is present for knet
                if transport in constants.TRANSPORTS_KNET:
                    raw_options[linknumber]["linknumber"] = linknumber
        return {
            linknumber: self.__translate_link_options(options, False)
            for linknumber, options in raw_options.items()
        }

    def get_transport(self):
        transport = self._get_option_value("totem", "transport")
        return transport if transport else constants.TRANSPORT_DEFAULT

    def get_ip_version(self):
        ip_version = self._get_option_value("totem", "ip_version")
        if ip_version:
            return ip_version
        if self.get_transport() == "udp":
            return constants.IP_VERSION_4
        return constants.IP_VERSION_64

    def _get_option_value(
        self, section: str, option: str, default: Optional[str] = None
    ) -> Optional[str]:
        for sec in self.config.get_sections(section):
            default = sec.get_attribute_value(option, default)
        return default

    def _is_changed(
        self, section: str, option_name: str, new_value: Optional[str]
    ) -> bool:
        if new_value is None:
            return False
        old_value = self._get_option_value(section, option_name)
        # old_value is not present or empty and new_value is empty
        if not old_value and not new_value:
            return False
        return old_value != new_value

    def set_transport_options(
        self,
        transport_options: Mapping[str, str],
        compression_options: Mapping[str, str],
        crypto_options: Mapping[str, str],
    ) -> None:
        """
        Set transport options for transport type currently used

        generic_options -- generic transport options
        compression_options -- compression options
        crypto_options -- crypto options
        """
        if any(
            self._is_changed("totem", opt, transport_options.get(opt))
            for opt in constants.TRANSPORT_RUNTIME_CHANGE_BANNED_OPTIONS
        ):
            self._need_stopped_cluster = True
        transport_type = self.get_transport()
        if transport_type in constants.TRANSPORTS_KNET:
            self._set_transport_knet_options(
                transport_options, compression_options, crypto_options
            )
        elif transport_type in constants.TRANSPORTS_UDP:
            self._set_transport_udp_options(transport_options)

    def _set_transport_udp_options(self, options):
        """
        Set transport options for udp transports

        dict options -- transport options
        """
        totem_section_list = self.__ensure_section(self.config, "totem")
        self.__set_section_options(totem_section_list, options)
        self.__remove_empty_sections(self.config)

    def _set_transport_knet_options(
        self, generic_options, compression_options, crypto_options
    ):
        """
        Set transport options for knet transport

        dict generic_options -- generic transport options
        dict compression_options -- compression options
        dict crypto_options -- crypto options
        """
        totem_section_list = self.__ensure_section(self.config, "totem")
        self.__set_section_options(totem_section_list, generic_options)
        self.__set_section_options(
            totem_section_list,
            _add_prefix_to_dict_keys(
                _KNET_COMPRESSION_OPTIONS_PREFIX, compression_options
            ),
        )
        self.__set_section_options(
            totem_section_list,
            _add_prefix_to_dict_keys(
                _KNET_CRYPTO_OPTIONS_PREFIX, crypto_options
            ),
        )
        self.__remove_empty_sections(self.config)

    def _filter_options(
        self,
        section_name: str,
        allowed_options: Iterable[str],
        prefix: str = "",
    ) -> Dict[str, str]:
        options = {}
        for section in self.config.get_sections(section_name):
            for name, value in section.get_attributes():
                if (
                    name.startswith(prefix)
                    and name[len(prefix) :] in allowed_options
                ):
                    options[name[len(prefix) :]] = value
        return options

    def get_transport_options(self) -> Dict[str, str]:
        """
        Get configurable generic transport options
        """
        transport_type = self.get_transport()
        if transport_type in constants.TRANSPORTS_KNET:
            return self._filter_options(
                "totem", constants.TRANSPORT_KNET_GENERIC_OPTIONS
            )
        if transport_type in constants.TRANSPORTS_UDP:
            return self._filter_options(
                "totem", constants.TRANSPORT_UDP_GENERIC_OPTIONS
            )
        return {}

    def get_compression_options(self) -> Dict[str, str]:
        """
        Get configurable compression options
        """
        return self._filter_options(
            "totem",
            constants.TRANSPORT_KNET_COMPRESSION_OPTIONS,
            _KNET_COMPRESSION_OPTIONS_PREFIX,
        )

    def get_crypto_options(self) -> Dict[str, str]:
        """
        Get configurable crypto options
        """
        return self._filter_options(
            "totem",
            constants.TRANSPORT_KNET_CRYPTO_OPTIONS,
            _KNET_CRYPTO_OPTIONS_PREFIX,
        )

    def set_totem_options(self, options):
        """
        Set options in the "totem" section

        dict options -- totem options
        """
        totem_section_list = self.__ensure_section(self.config, "totem")
        self.__set_section_options(totem_section_list, options)
        self.__remove_empty_sections(self.config)
        # The totem section contains quite a lot of options. Pcs reduces the
        # number by moving some of them virtually into other "sections". The
        # move is only visible for pcs users (in CLI, web UI), those moved
        # options are still written to the totem section. The options which pcs
        # keeps in the totem section for users do not currently need the
        # cluster to be stopped when updating them.
        # Note: all options in totem section supported by pcs are runtime
        # configurable.

    def get_totem_options(self) -> Dict[str, str]:
        """
        Get configurable totem options
        """
        return self._filter_options("totem", constants.TOTEM_OPTIONS)

    def set_quorum_options(self, options):
        """
        Set options in the "quorum" section

        dict options -- quorum options
        """
        quorum_section_list = self.__ensure_section(self.config, "quorum")
        self.__set_section_options(quorum_section_list, options)
        self.__update_two_node()
        self.__remove_empty_sections(self.config)
        self._need_stopped_cluster = True

    def get_quorum_options(self) -> Dict[str, str]:
        """
        Get configurable options from the "quorum" section
        """
        return self._filter_options("quorum", constants.QUORUM_OPTIONS)

    def is_enabled_auto_tie_breaker(self):
        """
        Returns True if auto tie braker option is enabled, False otherwise.
        """
        return self._get_option_value("quorum", "auto_tie_breaker", "0") == "1"

    def has_quorum_device(self):
        """
        Check if quorum device is present in the config
        """
        for quorum in self.config.get_sections("quorum"):
            for device in quorum.get_sections("device"):
                if device.get_attributes("model"):
                    return True
        return False

    def get_quorum_device_model(self):
        """
        Get quorum device model from quorum.device section
        """
        return self.get_quorum_device_settings()[0]

    def get_quorum_device_settings(self):
        """
        Get configurable options from quorum.device section
        """
        model = None
        model_options = {}
        generic_options = {}
        heuristics_options = {}
        for quorum in self.config.get_sections("quorum"):
            for device in quorum.get_sections("device"):
                for name, value in device.get_attributes():
                    if name == "model":
                        model = value
                    else:
                        generic_options[name] = value
                for subsection in device.get_sections():
                    if subsection.name == "heuristics":
                        heuristics_options.update(subsection.get_attributes())
                        continue
                    if subsection.name not in model_options:
                        model_options[subsection.name] = {}
                    model_options[subsection.name].update(
                        subsection.get_attributes()
                    )
        return (
            model,
            model_options.get(model, {}),
            generic_options,
            heuristics_options,
        )

    def is_quorum_device_heuristics_enabled_with_no_exec(self):
        heuristics_options = self.get_quorum_device_settings()[3]
        regexp = constants.QUORUM_DEVICE_HEURISTICS_EXEC_NAME_RE
        exec_found = False
        for name, value in heuristics_options.items():
            if value and regexp.match(name):
                exec_found = True
                break
        return not exec_found and heuristics_options.get("mode") in (
            "on",
            "sync",
        )

    def add_quorum_device(
        self, model, model_options, generic_options, heuristics_options
    ):
        # pylint: disable=too-many-locals
        """
        Add quorum device configuration

        string model -- quorum device model
        dict model_options -- model specific options
        dict generic_options -- generic quorum device options
        dict heuristics_options -- heuristics options
        """
        if self.has_quorum_device():
            raise LibraryError(
                ReportItem.error(reports.messages.QdeviceAlreadyDefined())
            )

        # configuration cleanup
        remove_need_stopped_cluster = {
            name: ""
            for name in constants.QUORUM_OPTIONS_INCOMPATIBLE_WITH_QDEVICE
        }
        # remove old device settings
        quorum_section_list = self.__ensure_section(self.config, "quorum")
        for quorum in quorum_section_list:
            for device in quorum.get_sections("device"):
                quorum.del_section(device)
            for name, value in quorum.get_attributes():
                if name in remove_need_stopped_cluster and value not in [
                    "",
                    "0",
                ]:
                    self._need_stopped_cluster = True
        # remove conflicting quorum options
        attrs_to_remove = {
            "allow_downscale": "",
            "two_node": "",
        }
        attrs_to_remove.update(remove_need_stopped_cluster)
        self.__set_section_options(quorum_section_list, attrs_to_remove)
        # remove nodes' votes
        for nodelist in self.config.get_sections("nodelist"):
            for node_section in nodelist.get_sections("node"):
                node_section.del_attributes_by_name("quorum_votes")

        # add new configuration
        quorum = quorum_section_list[-1]
        new_device = config_parser.Section("device")
        quorum.add_section(new_device)
        self.__set_section_options([new_device], generic_options)
        new_device.set_attribute("model", model)
        new_model = config_parser.Section(model)
        self.__set_section_options([new_model], model_options)
        new_device.add_section(new_model)
        new_heuristics = config_parser.Section("heuristics")
        self.__set_section_options([new_heuristics], heuristics_options)
        new_device.add_section(new_heuristics)

        self.__update_qdevice_votes()
        self.__update_two_node()
        self.__remove_empty_sections(self.config)

    def update_quorum_device(
        self, model_options, generic_options, heuristics_options
    ):
        """
        Update existing quorum device configuration

        dict model_options -- model specific options
        dict generic_options -- generic quorum device options
        dict heuristics_options -- heuristics options
        """
        if not self.has_quorum_device():
            raise LibraryError(
                ReportItem.error(reports.messages.QdeviceNotDefined())
            )
        model = self.get_quorum_device_model()

        # set new configuration
        device_sections = []
        model_sections = []
        heuristics_sections = []

        for quorum in self.config.get_sections("quorum"):
            device_sections.extend(quorum.get_sections("device"))
            for device in quorum.get_sections("device"):
                model_sections.extend(device.get_sections(model))
                heuristics_sections.extend(device.get_sections("heuristics"))
        # we know device sections exist, otherwise the function would exit at
        # has_quorum_device line above
        if not model_sections:
            new_model = config_parser.Section(model)
            device_sections[-1].add_section(new_model)
            model_sections.append(new_model)
        if not heuristics_sections:
            new_heuristics = config_parser.Section("heuristics")
            device_sections[-1].add_section(new_heuristics)
            heuristics_sections.append(new_heuristics)

        self.__set_section_options(device_sections, generic_options)
        self.__set_section_options(model_sections, model_options)
        self.__set_section_options(heuristics_sections, heuristics_options)

        self.__update_qdevice_votes()
        self.__update_two_node()
        self.__remove_empty_sections(self.config)
        self._need_qdevice_reload = True

    def remove_quorum_device_heuristics(self):
        """
        Remove quorum device heuristics configuration
        """
        if not self.has_quorum_device():
            raise LibraryError(
                ReportItem.error(reports.messages.QdeviceNotDefined())
            )
        for quorum in self.config.get_sections("quorum"):
            for device in quorum.get_sections("device"):
                for heuristics in device.get_sections("heuristics"):
                    device.del_section(heuristics)
        self.__remove_empty_sections(self.config)
        self._need_qdevice_reload = True

    def remove_quorum_device(self):
        """
        Remove all quorum device configuration
        """
        if not self.has_quorum_device():
            raise LibraryError(
                ReportItem.error(reports.messages.QdeviceNotDefined())
            )
        for quorum in self.config.get_sections("quorum"):
            for device in quorum.get_sections("device"):
                quorum.del_section(device)
        self.__update_two_node()
        self.__remove_empty_sections(self.config)

    def __update_two_node(self):
        # get relevant status
        has_quorum_device = self.has_quorum_device()
        has_two_nodes = len(self.get_nodes()) == 2
        auto_tie_breaker = self.is_enabled_auto_tie_breaker()
        # update two_node
        if has_two_nodes and not auto_tie_breaker and not has_quorum_device:
            quorum_section_list = self.__ensure_section(self.config, "quorum")
            self.__set_section_options(quorum_section_list, {"two_node": "1"})
        else:
            for quorum in self.config.get_sections("quorum"):
                quorum.del_attributes_by_name("two_node")

    def __update_qdevice_votes(self):
        # ffsplit won't start if votes is missing or not set to 1
        # for other algorithms it's required not to put votes at all
        model = None
        algorithm = None
        device_sections = []
        for quorum in self.config.get_sections("quorum"):
            for device in quorum.get_sections("device"):
                device_sections.append(device)
                model = device.get_attribute_value("model", model)
        for device in device_sections:
            for model_section in device.get_sections(model):
                algorithm = model_section.get_attribute_value(
                    "algorithm", algorithm
                )
        if model == "net":
            if algorithm == "ffsplit":
                self.__set_section_options(device_sections, {"votes": "1"})
            else:
                self.__set_section_options(device_sections, {"votes": ""})

    @staticmethod
    def __set_section_options(section_list, options):
        for section in section_list[:-1]:
            for name in options:
                section.del_attributes_by_name(name)
        for name, value in sorted(options.items()):
            if value == "":
                section_list[-1].del_attributes_by_name(name)
            else:
                section_list[-1].set_attribute(name, value)

    @staticmethod
    def __ensure_section(parent_section, section_name):
        section_list = parent_section.get_sections(section_name)
        if not section_list:
            new_section = config_parser.Section(section_name)
            parent_section.add_section(new_section)
            section_list.append(new_section)
        return section_list

    def __remove_empty_sections(self, parent_section):
        for section in parent_section.get_sections():
            self.__remove_empty_sections(section)
            if section.empty or (
                section.name == "interface"
                and list(section.get_attributes_dict().keys()) == ["linknumber"]
            ):
                parent_section.del_section(section)

    @staticmethod
    def __translate_link_options(options, input_to_corosync=True):
        pairs = constants.LINK_OPTIONS_KNET_TRANSLATION
        if input_to_corosync:
            translate_map = {pair[0]: pair[1] for pair in pairs}
        else:
            translate_map = {pair[1]: pair[0] for pair in pairs}
        result = {
            translate_map.get(name, name): value
            for name, value in options.items()
        }

        if "broadcast" in result:
            if input_to_corosync:
                # If broadcast == 1, transform it to broadcast == yes. If this
                # is called from an update where broadcast is being disabled,
                # remove broadcast from corosync.conf. Else do not put the
                # option to the config at all. From man corosync.conf, there is
                # only one allowed value: "yes".
                if result["broadcast"] in ("1", 1):
                    result["broadcast"] = "yes"
                elif result["broadcast"] in ("0", 0, ""):
                    result["broadcast"] = ""
                else:
                    del result["broadcast"]
            else:
                # When displaying config to users, do the opposite
                # transformation: only "yes" is allowed.
                if result["broadcast"] == "yes":
                    result["broadcast"] = "1"
                else:
                    del result["broadcast"]

        return result


def _add_prefix_to_dict_keys(prefix, data):
    return {"{}{}".format(prefix, key): value for key, value in data.items()}
